from __future__ import division

import numpy as np
from sklearn.externals.six import iteritems
from sklearn.externals.six.moves import xrange, zip


def _linear(Bs, dim, num_q, rhos, nus, clamp=True):
    # and the rest of the estimator is
    #   B / m * mean(nu ^ -dim)
    est = Bs / num_q * np.mean(nus ** (-dim), axis=0)
    if clamp:
        np.maximum(est, 0, out=est)
    return est


def kl(Ks, dim, num_q, rhos, nus, clamp=True):
    r'''
    Estimate the KL divergence between distributions:
        \int p(x) \log (p(x) / q(x))
    using the kNN-based estimator (5) of
        Qing Wang, Sanjeev R Kulkarni, and Sergio Verdu (2009).
        Divergence Estimation for Multidimensional Densities Via
        k-Nearest-Neighbor Distances.
        IEEE Transactions on Information Theory.
        http://www.ee.princeton.edu/~verdu/reprints/WanKulVer.May2009.pdf
    which is:
        d * 1/n \sum \log (nu_k(i) / rho_k(i)) + log(m / (n - 1))

    If clamp, enforces KL >= 0.

    Returns an array of shape (num_Ks,).
    '''
    est = dim * np.mean(np.log(nus) - np.log(rhos), axis=0)
    est += np.log(num_q / (rhos.shape[0] - 1))
    if clamp:
        np.maximum(est, 0, out=est)
    return est
kl.self_value = 0
kl.needs_alpha = False


def _alpha_div(omas, Bs, dim, num_q, rhos, nus, clamp=True):
    N = rhos.shape[0]

    # the actual main estimate:
    #   rho^(- dim * est alpha) nu^(- dim * est beta)
    #   = (rho / nu) ^ (dim * (1 - alpha))
    # do some reshaping trickery to get broadcasting right
    estimates = (rhos / nus)[:, np.newaxis, :]
    estimates = estimates ** (dim * omas.reshape(1, -1, 1))
    estimates = np.mean(estimates, axis=0)  # shape (n_alphas, n_Ks)

    estimates *= Bs

    # factors based on the sizes:
    #   1 / [ (n-1)^(est alpha) * m^(est beta) ] = ((n-1) / m) ^ (1 - alpha)
    estimates *= ((N - 1) / num_q) ** omas

    if clamp:
        np.maximum(estimates, 0, out=estimates)
    return estimates


def _jensen_shannon_core(Ks, dim, min_i, digamma_vals, num_q, rhos, nus, clamp=True):
    # We need to calculate the mean over points in X of
    # d * log radius of largest ball with no more than M/(n+m-1) weight
    #         where X points have weight 1 / (2 (n-1))
    #           and Y points have weight 1 / (2 m)
    # - digamma(# of neighbors in that ball)

    # ignores clamp.

    # NOTE: this is a stupidly slow implementation. the cython one should
    #       be much better, and also parallelize.
    num_p = rhos.shape[0]

    t = 2 * num_p - 1
    p_wt = 1 / t
    q_wt = num_p / (num_q * t)

    alphas = Ks / (num_p + num_q - 1)

    est = np.zeros(Ks.size)

    max_k = rhos.shape[1]
    combo = np.empty(max_k * 2, dtype=[('dist', np.float32), ('weight', float)])
    # could vectorize this loop if searchsorted worked axis-wise
    for rho, nu, in zip(rhos, nus):
        combo['dist'][:max_k] = rho
        combo['dist'][max_k:] = nu
        combo['weight'][:max_k] = p_wt
        combo['weight'][max_k:] = q_wt
        combo.sort()
        quantiles = np.cumsum(combo['weight'])

        i = quantiles.searchsorted(alphas, side='right')  # number pts in ball
        assert i.min() >= min_i

        est += dim * np.log(combo['dist'][i - 1]) - digamma_vals[i - min_i]
    est /= num_p
    return est


################################################################################

def _estimate_cross_divs(X_features, X_indices, X_rhos,
                         Y_features, Y_indices, Y_rhos,
                         funcs, Ks, max_K, save_all_Ks,
                         n_output, do_sym, to_self,
                         log_progress, n_jobs, min_dist, clamp):
    n_X = len(X_indices)
    n_Y = len(Y_features)

    K_indices = Ks - 1
    which_Ks = slice(None, None) if save_all_Ks else K_indices
    n_Ks = len(Ks)

    outputs = np.empty([n_output, n_Ks, n_X, n_Y, 2 if do_sym else 1],
                       dtype=np.float32)
    outputs.fill(np.nan)

    # Keep track of whether each function needs rho_sub or just rho
    if save_all_Ks:
        def needs_sub(func):
            return not getattr(func, 'needs_all_ks', False)
    else:
        def needs_sub(func):
            return False

    X_features.make_stacked()
    X_feats = X_features.stacked_features
    X_bounds = X_features._boundaries

    # rho: kth neighbors of each X in X
    # nu:  kth neighbors of each X in Y
    # so we can get the nus for all Xs into one Y at once (one column of out)

    Y_indices_loop = log_progress(Y_indices) if log_progress else Y_indices
    for j, Y_index in enumerate(Y_indices_loop):
        # Loop over columns of the output array,
        # searching / getting nus from all of the X bags to this one Y bag.

        # TODO: if X = Y, this wastes time for searches to self

        num_q = Y_features.n_pts[j]

        # KNN search is parallelized with openmp in flann
        knns = np.sqrt(Y_index.nn_index(X_feats, max_K)[1][:, which_Ks])
        np.maximum(min_dist, knns, out=knns)

        # all this is sequential. that's why this implementation is slow.
        for i in xrange(n_X):
            rho = X_rhos[i]
            nu = knns[X_bounds[i]:X_bounds[i + 1]]

            if save_all_Ks:
                rho_sub = rho[:, K_indices]
                nu_sub = nu[:, K_indices]

            for func, info in iteritems(funcs):
                o = (info.pos, slice(None), i, j, 0)

                if i == j and to_self and func.self_value is not None:
                    outputs[o] = func.self_value
                elif needs_sub(func):
                    outputs[o] = func(num_q, rho_sub, nu_sub, clamp=clamp)
                else:
                    outputs[o] = func(num_q, rho, nu, clamp=clamp)

    if do_sym:
        if X_features == Y_features:
            w = slice(None) if do_sym == True else list(do_sym)
            outputs[w, ..., 1] = outputs[w, ..., 0].transpose(0, 1, 3, 2)
            return outputs

        Y_features.make_stacked()
        Y_feats = Y_features.stacked_features
        Y_bounds = Y_features._boundaries

        if do_sym == True:
            sym_funcs = funcs
        else:
            sym_funcs = {func: info for func, info in iteritems(funcs)
                         if not do_sym.isdisjoint(info.pos)}

        X_indices_loop = log_progress(X_indices) if log_progress else X_indices
        for i, X_index in enumerate(X_indices_loop):
            # Loop over rows of the output array,
            # searching / getting nus from all of the Y bags to this one X bag.

            num_q = X_features.n_pts[i]

            knns = np.sqrt(X_index.nn_index(Y_feats, max_K)[1][:, which_Ks])
            np.maximum(min_dist, knns, out=knns)

            for j in xrange(n_Y):
                rho = Y_rhos[j]
                nu = knns[Y_bounds[j]:Y_bounds[j + 1]]

                if save_all_Ks:
                    rho_sub = rho[:, K_indices]
                    nu_sub = nu[:, K_indices]

                for func, info in iteritems(sym_funcs):
                    o = (info.pos, slice(None), i, j, 1)

                    if i == j and to_self and func.self_value is not None:
                        outputs[o] = func.self_value
                    elif needs_sub(func):
                        outputs[o] = func(num_q, rho_sub, nu_sub, clamp=clamp)
                    else:
                        outputs[o] = func(num_q, rho, nu, clamp=clamp)

    return outputs



from __future__ import division

from functools import partial
import logging
import os
import sys
import tempfile

import numpy as np
from scipy.special import psi
from sklearn.externals.six.moves import xrange, zip

from nose.tools import assert_raises
from sklearn.utils.testing import assert_array_almost_equal
from nose import SkipTest
from testfixtures import LogCapture

if __name__ == '__main__':
    # make this copy of skl_groups importable
    _this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(_this_dir)))

from skl_groups.features import Features


have_flann = True
try:
    import cyflann
except ImportError:
    try:
        import pyflann
    except ImportError:
        have_flann = False

if have_flann:
    from skl_groups.divergences import KNNDivergenceEstimator

    try:
        import skl_groups_accel.knn_divs
    except ImportError:
        have_accel = False
    else:
        have_accel = True


################################################################################

def test_knn_version_consistency():
    if not have_flann:
        raise SkipTest("No flann, so skipping knn tests.")
    if not have_accel:
        raise SkipTest("No skl-groups-accel, so skipping version consistency.")

    n = 20
    for dim in [1, 7]:
        np.random.seed(47)
        bags = Features([np.random.randn(np.random.randint(30, 100), dim)
                         for _ in xrange(n)])

        div_funcs = ('kl', 'js', 'renyi:.9', 'l2', 'tsallis:.8')
        Ks = (3, 4)
        get_est = partial(KNNDivergenceEstimator, div_funcs=div_funcs, Ks=Ks)
        results = {}
        for version in ('fast', 'slow', 'best'):
            est = get_est(version=version)
            results[version] = res = est.fit_transform(bags)
            assert res.shape == (len(div_funcs), len(Ks), n, n)
            assert np.all(np.isfinite(res))

        for df, fast, slow in zip(div_funcs, results['fast'], results['slow']):
            assert_array_almost_equal(
                fast, slow, decimal=1 if df == 'js' else 5,
                err_msg="({}, dim {})".format(df, dim))
            # TODO: debug JS differences

        est = get_est(version='fast', n_jobs=-1)
        res = est.fit_transform(bags)
        assert np.all(results['fast'] == res)

        est = get_est(version='slow', n_jobs=-1)
        res = est.fit_transform(bags)
        assert np.all(results['slow'] == res)


def test_knn_sanity_slow():
    if not have_flann:
        raise SkipTest("No flann, so skipping knn tests.")

    dim = 3
    n = 20
    np.random.seed(47)
    bags = Features([np.random.randn(np.random.randint(30, 100), dim)
                     for _ in xrange(n)])

    # just make sure it runs
    div_funcs = ('kl', 'js', 'renyi:.9', 'l2', 'tsallis:.8')
    Ks = (3, 4)
    est = KNNDivergenceEstimator(div_funcs=div_funcs, Ks=Ks)
    res = est.fit_transform(bags)
    assert res.shape == (len(div_funcs), len(Ks), n, n)
    assert np.all(np.isfinite(res))

    # test that JS blows up when there's a huge difference in bag sizes
    # (so that K is too low)
    assert_raises(
        ValueError,
        partial(est.fit_transform, bags + [np.random.randn(1000, dim)]))

    # test fit() and then transform() with JS, with different-sized test bags
    est = KNNDivergenceEstimator(div_funcs=('js',), Ks=(5,))
    est.fit(bags, get_rhos=True)
    with LogCapture('skl_groups.divergences.knn', level=logging.WARNING) as l:
        res = est.transform([np.random.randn(300, dim)])
        assert res.shape == (1, 1, 1, len(bags))
        assert len(l.records) == 1
        assert l.records[0].message.startswith('Y_rhos had a lower max_K')

    # test that passing div func more than once raises
    def blah(df):
        est = KNNDivergenceEstimator(div_funcs=[df, df])
        return est.fit(bags)
    assert_raises(ValueError, lambda: blah('kl'))
    assert_raises(ValueError, lambda: blah('renyi:.8'))
    assert_raises(ValueError, lambda: blah('l2'))


def test_knn_memory():
    if not have_flann:
        raise SkipTest("No flann, so skipping knn tests.")

    dim = 3
    n = 20
    np.random.seed(47)
    bags = Features([np.random.randn(np.random.randint(30, 100), dim)
                     for _ in xrange(n)])

    tdir = tempfile.mkdtemp()
    div_funcs = ('kl', 'js', 'renyi:.9', 'l2', 'tsallis:.8')
    Ks = (3, 4)
    est = KNNDivergenceEstimator(div_funcs=div_funcs, Ks=Ks, memory=tdir)
    res1 = est.fit_transform(bags)

    with LogCapture('skl_groups.divergences.knn', level=logging.INFO) as l:
        res2 = est.transform(bags)
        assert len(l.records) == 0
    assert np.all(res1 == res2)

    with LogCapture('skl_groups.divergences.knn', level=logging.INFO) as l:
        res3 = est.fit_transform(bags)
        for r in l.records:
            assert not r.message.startswith("Getting divergences")
    assert np.all(res1 == res3)


def test_knn_kl():
    if not have_flann:
        raise SkipTest("No flann, so skipping knn tests.")

    # verified by hand
    # Dhat(P||Q) = \log m/(n-1) + d / n  \sum_{i=1}^n \log \nu_k(i)/rho_k(i)
    x = np.reshape([0., 1, 3], (3, 1))
    y = np.reshape([.2, 1.2, 3.2, 7.2], (4, 1))

    n = x.shape[0]
    m = y.shape[0]

    x_to_y = np.log(m / (n-1)) + 1/n * (
        np.log(1.2 / 3) + np.log(.8 / 2) + np.log(1.8 / 3))
    y_to_x = np.log(n / (m-1)) + 1/m * (
        np.log(.8 / 3) + np.log(1.2 / 2) + np.log(2.2 / 3) + np.log(6.2 / 6))

    msg = "got {}, expected {}"
    est = KNNDivergenceEstimator(div_funcs=['kl'], Ks=[2], clamp=False)
    res = est.fit_transform([x, y]).squeeze()
    assert res[0, 0] == 0
    assert res[1, 1] == 0
    assert np.allclose(res[0, 1], x_to_y), msg.format(res[0, 1], x_to_y)
    assert np.allclose(res[1, 0], y_to_x), msg.format(res[1, 0], y_to_x)


def test_knn_js():
    if not have_flann:
        raise SkipTest("No flann, so skipping knn tests.")

    # verified by hand
    x = np.reshape([0., 1, 3, 6], (4, 1))
    n = 4

    y = np.reshape([.2, 1.2, 3.2, 6.2, 10.2], (5, 1))
    m = 5

    M = 2

    right_js = (
        np.log(n + m - 1) + psi(M)
        + 1/(2*n) * (  # x weight is 1/7, y weight is 4/35, quantile 1/4
            np.log(.2) - psi(1)     # 0 => .2(y), 1(x)
            + np.log(.8) - psi(2)   # 1 => 1.2(y), .2(y)
            + np.log(1.8) - psi(2)  # 3 => 3.2(y), 1.2(y)
            + np.log(2.8) - psi(2)  # 6 => 6.2(y), 3.2(y)
        )
        + 1/(2*m) * (  # x weight is 5/36, y weight is 1/9, quantile 1/4
            np.log(.2) - psi(1)     # .2 => 0(x)
            + np.log(1) - psi(2)    # 1.2 => 1(x), .2(y)
            + np.log(2) - psi(2)    # 3.2 => 3(x), 1.2(y)
            + np.log(3) - psi(2)    # 6.2 => 6(x), 3.2(y)
            + np.log(4.2) - psi(2)  # 10.2 => 6.2(y), 6(x)
        )
        - 1/2 * np.log(n-1) - 1/(2*n) * (
            np.log(3) + np.log(2) + np.log(3) + np.log(5))
        - 1/2 * np.log(m-1) - 1/(2*m) * (
            np.log(3) + np.log(2) + np.log(3) + np.log(4) + np.log(7))
    )

    msg = "got {}, expected {}"
    est = KNNDivergenceEstimator(div_funcs=['js'], Ks=[2], clamp=False)
    res = est.fit([x]).transform([y])
    assert res.shape == (1, 1, 1, 1)
    res = res[0, 0, 0, 0]
    assert np.allclose(res, right_js, atol=1e-6), msg.format(res, right_js)



################################################################################

if __name__ == '__main__':
    import nose
    nose.main()



from __future__ import division

import os
import sys

import numpy as np
from sklearn.externals.six import iteritems
from nose.tools import assert_raises

if __name__ == '__main__':
    # make this copy of skl_groups importable
    _this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(_this_dir)))

from skl_groups import utils


################################################################################

def test_type_utils():
    tests = {
        'bool': (np.array([False, True]), False, True),
        'int32': (np.arange(10, dtype=np.int32), True, True),
        'int64': (np.arange(10, dtype=np.int64), True, True),
        'float32': (np.arange(10, dtype=np.float32), False, False),
        'float64': (np.arange(10, dtype=np.float64), False, False),
    }

    for name, (a, is_int, is_cat) in iteritems(tests):
        assert utils.is_integer_type(a) == is_int, name
        assert utils.is_categorical_type(a) == is_cat, name

        assert utils.is_integer(a[0]) == is_int, name
        assert utils.is_categorical(a[0]) == is_cat, name

    assert utils.is_integer_type(utils.as_integer_type(tests['float32'][0]))
    assert utils.is_integer_type(utils.as_integer_type(tests['float64'][0]))
    assert_raises(
        ValueError, lambda: utils.as_integer_type(tests['float32'][0] + .2))

    assert utils.is_integer(5)
    assert utils.is_categorical(False)
    assert utils.is_categorical(True)

################################################################################

if __name__ == '__main__':
    import nose
    nose.main()


from __future__ import division

import os
import sys

import numpy as np
from sklearn.metrics.pairwise import euclidean_distances
from nose.tools import assert_raises

if __name__ == '__main__':
    # make this copy of skl_groups importable
    _this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(_this_dir)))

from skl_groups.kernels.transform import (
        PairwisePicker, Symmetrize, RBFize,
        ProjectPSD, FlipPSD, ShiftPSD, SquarePSD)

################################################################################

def test_pairwise_picker():
    X = np.arange(64).reshape((2, 2, 4, 4))
    pick = PairwisePicker((0, 1))
    Xt = pick.fit_transform(X)
    assert np.all(Xt == np.arange(16, 32).reshape((4, 4)))

    pick = PairwisePicker[:, 0]
    assert np.all(pick.fit_transform(X) == X[:, 0])


def test_symmetrize():
    X = np.random.normal(size=(10, 10))
    Xo = X.copy()

    sym = Symmetrize(copy=True)
    Xs = sym.fit_transform(X)
    assert np.all(X == Xo)
    assert np.allclose(Xs, (X + X.T) / 2)
    assert np.all(sym.transform(X + 1) == X + 1)

    Xs2 = Symmetrize(copy=False).fit_transform(X)
    assert np.allclose(Xs2, (X + X.T) / 2)

    sym = Symmetrize()
    assert_raises(NotImplementedError, lambda: sym.fit(X))
    assert_raises(TypeError, lambda: sym.fit_transform(np.zeros((5, 3))))


def test_rbfize():
    X = np.random.normal(size=(20, 4))
    dists = euclidean_distances(X)
    median = np.median(dists[np.triu_indices_from(dists, k=1)])

    rbf = RBFize(gamma=.25)
    res = rbf.fit_transform(dists)
    assert not hasattr(res, 'median_')
    assert np.allclose(res, np.exp(-.25 * dists ** 2))

    rbf = RBFize(gamma=.25, squared=True)
    res = rbf.fit_transform(dists)
    assert np.allclose(res, np.exp(-.25 * dists))

    rbf = RBFize(gamma=4, scale_by_median=True)
    res = rbf.fit_transform(dists)
    assert np.allclose(rbf.median_, median)
    assert np.allclose(res, np.exp((-4 * median**2) * dists ** 2))

    rbf = RBFize(gamma=4, scale_by_median=True, squared=True)
    res = rbf.fit_transform(dists)
    assert np.allclose(rbf.median_, median)
    assert np.allclose(res, np.exp((-4 * median) * dists))



def test_project():
    X = np.diagflat([-2, -1, 0, 1, 2])
    # eigenvalues -2, -1, 0, 1, 2; eigenvectors are I

    Xproj = ProjectPSD().fit_transform(X)
    assert np.allclose(Xproj, np.diagflat([0, 0, 0, 1, 2]))

    Xproj2 = ProjectPSD().fit(X).transform(X)
    assert np.allclose(Xproj2, np.diagflat([0, 0, 0, 1, 2]))

    Xproj3 = ProjectPSD(negatives_likely=True).fit(X).transform(X[:3, :])
    assert np.allclose(Xproj3, np.zeros((3, 5)))

    Xproj4 = ProjectPSD(negatives_likely=False).fit(X).transform(X[:3, :])
    assert np.allclose(Xproj4, np.zeros((3, 5)))

    Xproj5 = ProjectPSD(negatives_likely=True, copy=False, min_eig=.5) \
            .fit_transform(X.copy())
    assert np.allclose(Xproj5, np.diagflat([.5, .5, .5, 1, 2]))

    Xproj6 = ProjectPSD(negatives_likely=True, copy=False, min_eig=.5) \
            .fit(X.copy()).transform(X.copy())
    assert np.allclose(Xproj6, np.diagflat([.5, .5, 0, 1, 2]))

    assert_raises(TypeError, lambda: ProjectPSD().fit(X[:2, :]))
    assert_raises(TypeError, lambda: ProjectPSD().fit_transform(X[:2, :]))
    assert_raises(TypeError, lambda: ProjectPSD().fit(X).transform(X[:, :2]))


def test_flip():
    X = np.diagflat([-2, -1, 0, 1, 2])
    # eigenvalues -2, -1, 0, 1, 2; eigenvectors are I

    Xflip = FlipPSD().fit_transform(X)
    assert np.allclose(Xflip, np.diagflat([2, 1, 0, 1, 2]))

    Xflip2 = FlipPSD().fit(X).transform(X)
    assert np.allclose(Xflip2, np.diagflat([2, 1, 0, 1, 2]))

    Xflip3 = FlipPSD().fit(X).transform(X[:3, :])
    assert np.allclose(Xflip3, [[2,0,0,0,0], [0,1,0,0,0], [0,0,0,0,0]])

    assert_raises(TypeError, lambda: FlipPSD().fit(X[:2, :]))
    assert_raises(TypeError, lambda: FlipPSD().fit_transform(X[:2, :]))
    assert_raises(TypeError, lambda: FlipPSD().fit(X).transform(X[:, :2]))


def test_shift():
    X = np.diagflat([-2., -1, 0, 1, 2])
    # eigenvalues -2, -1, 0, 1, 2; eigenvectors are I

    Xshift = ShiftPSD().fit_transform(X)
    assert np.allclose(Xshift, np.diagflat([0, 1, 2, 3, 4]))

    Xshift2 = ShiftPSD().fit(X).transform(X)
    assert np.allclose(Xshift2, np.diagflat([0, 1, 2, 3, 4]))

    Xshift3 = ShiftPSD().fit(X).transform(X[:3, :])
    assert np.allclose(Xshift3, X[:3, :])

    Xshift4 = ShiftPSD(min_eig=2).fit_transform(X)
    assert np.allclose(Xshift4, np.diagflat([2, 3, 4, 5, 6]))

    assert_raises(TypeError, lambda: ShiftPSD().fit(X[:2, :]))
    assert_raises(TypeError, lambda: ShiftPSD().fit_transform(X[:2, :]))
    assert_raises(TypeError, lambda: ShiftPSD().fit(X).transform(X[:, :2]))


def test_square():
    X = np.diagflat([-2., -1, 0, 1, 2])
    # eigenvalues -2, -1, 0, 1, 2; eigenvectors are I

    Xsquare = SquarePSD().fit_transform(X)
    assert np.allclose(Xsquare, np.diagflat([4, 1, 0, 1, 4]))

    assert_raises(NotImplementedError, lambda: SquarePSD().fit(X))
    assert_raises(NotImplementedError, lambda: SquarePSD().transform(X))
    assert_raises(TypeError, lambda: SquarePSD().fit_transform(X[:2, :]))


################################################################################

if __name__ == '__main__':
    import nose
    nose.main()



from __future__ import division

from copy import copy, deepcopy
from functools import partial
import os
import sys
import warnings

import numpy as np
from sklearn.externals.six.moves import xrange, cPickle as pickle
from nose.tools import assert_raises

if __name__ == '__main__':
    # make this copy of skl_groups importable
    _this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(_this_dir)))

from skl_groups.features import Features


wrong_type = partial(assert_raises, (TypeError, IndexError))


################################################################################

def test_features_basic():
    bags = [np.random.normal(size=(np.random.randint(10, 100), 10))
            for _ in xrange(20)]

    assert repr(Features([[[8, 9], [12, 12]]]))

    feats = Features(bags, copy=False, stack=False, label=np.arange(20))
    assert len(feats) == 20
    assert feats.total_points == sum(bag.shape[0] for bag in bags)
    assert np.all(feats[3] == bags[3])
    assert np.all(feats.label == np.arange(20))
    assert repr(feats)
    assert feats.dtype == np.float64
    assert feats != bags
    assert feats.bare() == bags

    sub = feats[[5, 2]]
    assert np.all(sub.label == [5, 2])
    assert np.all(sub[0] == feats[5])
    assert np.all(sub[1] == feats[2])
    assert repr(sub)

    feats[4][0, 0] = 1000
    assert bags[4][0, 0] == 1000

    feats.make_stacked()
    assert feats != bags
    assert feats.bare() == bags
    assert len(feats) == 20
    assert feats.total_points == sum(bag.shape[0] for bag in bags)
    assert np.all(feats[3] == bags[3])
    assert np.all(feats.label == np.arange(20))
    feats[0][0, 0] = -800
    assert feats.features[0][0, 0] == -800
    assert feats.stacked_features[0, 0] == -800
    assert repr(feats)

    wrong_type(lambda: feats['label'])
    wrong_type(lambda: feats[['label']])
    wrong_type(lambda: feats[[3, 'label']])

    cop = feats.copy()
    assert cop == feats
    assert cop.stacked
    cop.make_stacked()
    assert cop == feats
    cop[0][0, 0] = 12
    assert cop != feats
    assert repr(cop)

    fs = lambda *a, **kw: partial(Features, *a, **kw)

    bags = np.random.normal(size=(10, 5))
    wrong_type(fs(bags))
    wrong_type(fs(bags, [[4], [12]]))
    wrong_type(fs(bags, []))
    wrong_type(fs(bags, [-3, 13]))
    wrong_type(fs(bags, [7.5, 2.5]))
    wrong_type(fs(bags, [7, 2]))
    wrong_type(fs(np.zeros((10, 0)), [7, 3]))

    bags = [np.random.normal(size=(5, 8)), np.random.normal(size=(6, 8))]
    wrong_type(fs(bags, [5, 6]))
    assert np.all(
        Features([[5, 6], [[7, 9], [0, 0]]])[0] == np.reshape([5, 6], (1, 2)))
    wrong_type(fs([ [[[5]]] ]))
    wrong_type(fs([["hello", "there"]]))
    wrong_type(fs([[np.arange(10, dtype=int)], [np.arange(10, dtype=float)]]))
    wrong_type(fs([np.random.randn(8, 7), np.random.randn(0, 7)]))

    assert np.all(
        Features([[[1, 2]], [[3, 4]]], stack=True).stacked_features
        == [[1, 2], [3, 4]])

    assert_raises(ValueError, fs(bags, labels=np.arange(3)))

    with warnings.catch_warnings(record=True) as w:
        Features(bags, total_points=[1, 2])
        assert len(w) == 1


def test_copy_constructor():
    bags = [np.random.normal(size=(np.random.randint(10, 100), 10))
            for _ in xrange(20)]
    unstacked = Features(bags, label=np.arange(20))

    oth_unstacked = Features(unstacked)
    assert oth_unstacked.label is unstacked.label
    assert oth_unstacked.features[0] is unstacked.features[0]
    assert oth_unstacked == unstacked

    oth_unstacked_bare = Features(unstacked, bare=True)
    assert oth_unstacked_bare == bags
    assert oth_unstacked_bare.bare() == oth_unstacked_bare

    oth_unstacked = Features(unstacked, label=np.ones(20))
    assert np.all(oth_unstacked.label == 1)

    oth_unstacked2 = Features(unstacked, bare=True, label=np.arange(20))
    assert oth_unstacked2 == unstacked

    oth_unstacked_copy = Features(unstacked, copy=True)
    assert oth_unstacked_copy == unstacked
    assert not np.may_share_memory(oth_unstacked_copy.features[0],
                                   unstacked.features[0])

    stacked = unstacked.copy()
    stacked.make_stacked()
    oth_stacked = Features(stacked)
    assert oth_stacked == stacked


def test_copying():
    bags = [np.random.normal(size=(np.random.randint(10, 100), 10))
            for _ in xrange(20)]

    unstacked = Features(bags, copy=False, stack=False, label=np.arange(20))
    stacked = Features(bags, stack=True, label=np.arange(20))

    assert unstacked == stacked

    unstacked_copy = copy(unstacked)
    assert not unstacked_copy.stacked
    assert stacked == unstacked_copy == unstacked
    assert unstacked_copy.label is unstacked.label
    assert not np.may_share_memory(unstacked[0], unstacked_copy[0])

    unstacked_deep = deepcopy(unstacked)
    assert not unstacked_deep.stacked
    assert stacked == unstacked_deep == unstacked
    assert unstacked_deep.label is not unstacked.label

    stacked_copy = copy(stacked)
    assert stacked_copy.stacked
    assert stacked == stacked_copy == unstacked
    assert stacked_copy.label is stacked.label

    stacked_deep = deepcopy(stacked)
    assert stacked_deep.stacked
    assert stacked == stacked_deep == unstacked
    assert stacked_deep.label is not stacked.label

    unstacked_stacked = unstacked.copy(stack=True)
    assert unstacked_stacked.stacked
    assert stacked == unstacked_stacked == stacked
    assert unstacked_stacked.label is unstacked.label

    unstacked_pickled = pickle.loads(pickle.dumps(unstacked))
    assert unstacked == unstacked_pickled
    assert not unstacked_pickled.stacked
    assert unstacked_pickled.label is not unstacked.label

    stacked_pickled = pickle.loads(pickle.dumps(stacked))
    assert stacked == stacked_pickled
    assert stacked_pickled.stacked
    assert stacked_pickled.label is not stacked.label


def test_feats_add():
    bags = [np.random.normal(size=(np.random.randint(10, 100), 10))
            for _ in xrange(20)]
    labels = np.arange(20)

    first_15 = Features(bags[:15], labels=labels[:15])
    last_5 = Features(bags[15:], labels=labels[15:])

    plus = first_15 + last_5
    assert len(plus) == 20
    assert plus[:15] == first_15

    plus_list = first_15 + bags[15:]
    assert len(plus_list) == 20
    assert not plus_list.meta
    assert np.all(plus_list[16] == bags[16])

    plus_singlelist = first_15 + [bags[18]]
    assert np.all(plus_singlelist[15] == bags[18])

    rplus_list = bags[15:] + first_15
    assert np.all(rplus_list[0] == bags[15])

    rplus_singlelist = [bags[15]] + first_15
    assert np.all(rplus_singlelist[0] == bags[15])
    assert rplus_singlelist[1:] == first_15.bare()

    wrong_type(lambda: first_15 + 12)
    wrong_type(lambda: 12 + first_15)

    assert_raises(ValueError, lambda: first_15 + np.asarray(bags))

################################################################################

if __name__ == '__main__':
    import nose
    nose.main()



from __future__ import division

from collections import namedtuple, defaultdict, OrderedDict
from functools import partial
import itertools
import logging
import warnings

import numpy as np
from scipy.special import gamma, gammaln, psi
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.externals.joblib import Memory
from sklearn.externals.six import iteritems, itervalues, string_types
from sklearn.externals.six.moves import map, zip, xrange

try:
    from cyflann import FLANNIndex, FLANNParameters
except ImportError:
    from pyflann import FLANN as FLANNIndex, FLANNParameters

from ..features import as_features
from ..utils import identity, ProgressLogger, as_integer_type
from ._knn import _linear, kl, _alpha_div, _jensen_shannon_core

from ._knn import _estimate_cross_divs
try:
    from skl_groups_accel.knn_divs import (
            _estimate_cross_divs as _estimate_cross_divs_fast)
    have_fast_version = True
except ImportError as e:
    have_fast_version = False
    fast_version_error = e


__all__ = ['KNNDivergenceEstimator']


################################################################################
### Logging setup
logger = logging.getLogger(__name__)

progress_logger = logging.getLogger(__name__ + '.progress')
progress_logger.propagate = False
progress_logger.addHandler(logging.NullHandler())

def plog(it, name=None, total=None):
    return ProgressLogger(progress_logger, name=name)(it, total=total)


################################################################################
### Main class

class KNNDivergenceEstimator(BaseEstimator, TransformerMixin):
    r'''
    Estimates various divergence functions between bags.

    Assumes that each bag represents an independent and identically distributed
    sample from some unknown probability distribution (on which certain
    technical assumptions are made), and estimates various distances between
    them. The valid divergence functions are, where p and q refer to the density
    functions for two bags:

    * 'kl': The `Kullback-Liebler divergence <https://en.wikipedia.org/wiki/KL_divergence>`_,
      which acts like a distance and has some attractive information-theoretic
      properties as well as often giving good results on machine learning
      problems in practice. Values are nonnegative but can go to infinity.
      :math:`\int p(x) \log\left( \frac{p(x)}{q(x)} \right) dx`.
      Estimated as in [1]_.
    * 'renyi:.8' or any other number (other than 1): The
      `Renyi-alpha divergence <https://en.wikipedia.org/wiki/Renyi_divergence#R.C3.A9nyi_divergence>`_,
      :math:`\frac{1}{\alpha - 1} \log \int p(x) \left( \frac{p(x)}{q(x)} \right)^{\alpha - 1} dx`.
      Converges to 'kl' as :math:`\alpha` goes to 1.
      Values are nonnegative, but can go to infinity.
      Estimated as in [2]_.
    * 'tsallis:.8' or any other number (other than 1): The 
      `Tsallis-alpha divergence <https://en.wikipedia.org/wiki/Tsallis_entropy>`_,
      :math:`\frac{1}{\alpha - 1} \left( \int p(x) \left(\frac{p(x)}{q(x)} \right)^{\alpha - 1} dx - 1 \right)`.
      Converges to `kl` as :math:`\alpha` goes to 1.
      Values are nonnegative, but can go to infinity.
      Estimated as in [2]_.
    * 'hellinger': The `Hellinger distance <https://en.wikipedia.org/wiki/Hellinger_distance>`_,
      which is a true distance (i.e. it is symmetric and its true value
      satisfies the triangle inequality) and is between 0 and 1.
      Defined by :math:`\sqrt{1 - \int \sqrt{p(x) q(x)} dx}`.
      Estimated as in [2]_.
    * 'bc': The `Bhattacharyya coefficient <https://en.wikipedia.org/wiki/Bhattacharyya_coefficient>`_,
      :math:`\int \sqrt{p(x) q(x)} dx`.
      This is an affinity rather than a distance and is between 0 and 1.
      Estimated as in [2]_.
    * 'l2': The :math:`L_2` distance between density functions
      :math:`\sqrt{ \int (p(x) - q(x))^2 dx }`.
      Estimated as in [2]_.
      A true distance function (symmetric and satisfies the triangle inequality.)
    * 'linear': :math:`\int p(x) q(x) dx`.
      Estimated as in [2]_.
    * 'jensen-shannon' or 'js': The
      `Jensen-Shannon divergence <https://en.wikipedia.org/wiki/Jensen%E2%80%93Shannon_divergence>`_.
      Equal to :math:`\frac{1}{2} D(p \| M) + \tfrac{1}{2} D(q \| M)`,
      where D is the Kullback-Liebler divergence
      and M is an equal mixture between p and q,
      or equivalently :math:`H[M] - \frac{1}{2} H[p] - \frac{1}{2} H[q]`
      where H is the Shannon entropy.
      Symmetric, its square root satisfies the triangle inequality,
      and is between 0 and :math:`\ln 2`.
      Estimated using [3]_.

    Note that even for the true distance functions,
    resulting kernels will not necessarily be positive semidefinite
    due to estimation error.

    Parameters
    ----------

    div_funcs : sequence of strings, optional, default ['kl']
        A sequence of divergence function spec strings, as above.
        For Renyi or Tsallis divergences, you can pass multiple values of
        alpha, e.g. ``['renyi:.9', 'renyi:.8', 'tsallis:.99']``.

    Ks : sequence of positive integers, optional, default [3]
        The Ks to use for the K-nearest-neighbor estimator.
        If you have very small bags (sizes less than, say, 25), try 1 or 2;
        if large, say more than 1000, try 4 or 5.
        Must be less than the smallest bag size; for the proof of [2]_ to work,
        should be at least 3 for most divergence functions.
        Jensen-Shannon treats this parameter a little differently.

    do_sym : boolean, optional, default False
        As well as returning D(X || Y), return D(Y || X).

    n_jobs : integer, optional, default 1
        The number of CPUs to use in the computation. -1 means 'all CPUs'.

    clamp : boolean, optional, default True
        For functions with bounded outputs, "clamp" them to lie within that
        range; for example, if the estimator for KL divergence gives a negative
        value, return 0 instead.

    min_dist : float, optional, default 1e-3
        Protect against nearly-identical points by treating any distances less
        than this amount as this number. Tiny distances screw up the estimators
        because they assume that the inputs are from a continuous distribution,
        where this doesn't happen.

    flann_algorithm : string, optional, default 'auto'
        Which algorithm to use in FLANN for nearest neighbors. Defaults to
        'auto', which chooses 'kdtree_single' if the input dimension is at most
        5 and 'linear' otherwise. In high-ish dimensional spaces, you can get
        much better performance at the cost of approximate answers by using
        other index types; see the FLANN documentation.

    flann_args : dictionary, optional, default {}
        Other arguments to pass to FLANN.

    version : one of {'best', 'fast', 'slow'}, optional, default 'best'
        Whether to use the fast Cython implementation from skl-groups-accel
        or the slower pure-Python implementation. 'best' chooses 'fast' if
        available and you aren't using custom divergence functions, 'slow'
        otherwise.

    memory : Instance of :class:`joblib.Memory` or string, optional
        Used to cache the indices and the output of :meth:`transform`.
        By default, no caching is done. If a string is given, it is the
        path to the caching directory.

    Attributes
    ----------

    `features_` : :class:`skl_groups.features.Features`
        The features passed to `fit` (except made :meth:`sklearn.features.Features.bare`).

    `indices_` : list of :class:`cyflann.FLANNIndex` or :class:`pyflann.FLANN`
        A FLANN index for each bag in `features_`.

    `rhos_` : list of arrays
        For each bag in `features_`, the distance to the Kth nearest neighbor
        of each point amongst its own bag.
        ``rhos_[i][j, k]`` is the distance to either the ``Ks[k]`` th or the
        ``k+1`` th nearest neighbor of ``features_[i][j, :]`` in
        ``features_[i]`` (not including ``features_[i][j, :]`` itself).
        It's the ``k+1`` th if Jensen-Shannon divergence is requested.

        May or may not be present after :meth:`fit`; will be after
        :meth:`transform`.

    See also
    --------
    :class:`skl_groups.kernels.PairwisePicker`
        Chooses a particular divergence function / K combination in a pipeline.
        Doing parameter tuning with :class:`KNNDivergenceEstimator` with a
        :attr:`memory` and then :class:`PairwisePicker` is more efficient
        than directly tuning over :class:`KNNDivergenceEstimator` arguments.


    Notes
    -----
    The convergence proof in [1]_ seems incorrect. The estimator seems to work
    well, in practice, though.

    The Jensen-Shannon estimation is performed by using the estimator of
    [3]_ to get :math:`H[\frac{1}{2}(X + Y)]`, by combining the two samples
    and assigning weights such that the total weight from each sample is equal,
    and subtracting the mean of :math:`H[X]` and :math:`H[Y]` according to the
    same estimator (with equally-weighted points). The ``K`` parameter is
    used for the value of :math:`M = n \alpha` in [3]_.

    References
    ----------
    .. [1] Q. Wang, S. Kulkarni, & S. Verdu (2009).
           Divergence Estimation for Multidimensional Densities Via
           k-Nearest-Neighbor Distances.
           IEEE Transactions on Information Theory, 55(5), 2392-2405.
    .. [2] B. Poczos, L. Xiong, D. J. Sutherland, & J. Schneider (2012).
           Nonparametric kernel estimators for image classification.
           In Computer Vision and Pattern Recognition (CVPR).
    .. [3] H. Hino & N. Murata (2013).
           Information estimators for weighted observations.
           Neural Networks.
    '''
    def __init__(self, div_funcs=('kl',), Ks=(3,), do_sym=False, n_jobs=1,
                 clamp=True, min_dist=1e-3,
                 flann_algorithm='auto', flann_args=None, version='best',
                 memory=Memory(cachedir=None, verbose=0)):
        self.div_funcs = div_funcs
        self.Ks = Ks
        self.do_sym = do_sym
        self.n_jobs = n_jobs
        self.clamp = clamp
        self.min_dist = min_dist
        self.flann_algorithm = flann_algorithm
        self.flann_args = flann_args
        self.version = version
        self.memory = memory

    def _get_Ks(self):
        "Ks as an array and type-checked."
        Ks = as_integer_type(self.Ks)
        if Ks.ndim != 1:
            raise TypeError("Ks should be 1-dim, got shape {}".format(Ks.shape))
        if Ks.min() < 1:
            raise ValueError("Ks should be positive; got {}".format(Ks.min()))
        return Ks

    @property
    def _n_jobs(self):
        "n_jobs, but with the number of cores instead of -1."
        if self.n_jobs == -1:
            from multiprocessing import cpu_count
            return cpu_count()
        return self.n_jobs

    def _flann_args(self, X=None):
        "The dictionary of arguments to give to FLANN."
        args = {'cores': self._n_jobs}
        if self.flann_algorithm == 'auto':
            if X is None or X.dim > 5:
                args['algorithm'] = 'linear'
            else:
                args['algorithm'] = 'kdtree_single'
        else:
            args['algorithm'] = self.flann_algorithm
        if self.flann_args:
            args.update(self.flann_args)

        # check that arguments are correct
        try:
            FLANNParameters().update(args)
        except AttributeError as e:
            msg = "flann_args contains an invalid argument:\n  {}"
            raise TypeError(msg.format(e))

        return args

    def fit(self, X, y=None, get_rhos=False):
        '''
        Sets up for divergence estimation "from" new data "to" X.
        Builds FLANN indices for each bag, and maybe gets within-bag distances.

        Parameters
        ----------
        X : list of arrays or :class:`skl_groups.features.Features`
            The bags to search "to".

        get_rhos : boolean, optional, default False
            Compute within-bag distances :attr:`rhos_`. These are only needed
            for some divergence functions or if do_sym is passed, and they'll
            be computed (and saved) during :meth:`transform` if they're not
            computed here.

            If you're using Jensen-Shannon divergence, a higher max_K may
            be needed once it sees the number of points in the transformed bags,
            so the computation here might be wasted.
        '''
        self.features_ = X = as_features(X, stack=True, bare=True)

        # if we're using a function that needs to pick its K vals itself,
        # then we need to set max_K here. when we transform(), might have to
        # re-do this :|
        Ks = self._get_Ks()
        _, _, _, max_K, save_all_Ks, _ = _choose_funcs(
            self.div_funcs, Ks, X.dim, X.n_pts, None, self.version)

        if max_K >= X.n_pts.min():
            msg = "asked for K = {}, but there's a bag with only {} points"
            raise ValueError(msg.format(max_K, X.n_pts.min()))

        memory = self.memory
        if isinstance(memory, string_types):
            memory = Memory(cachedir=memory, verbose=0)

        self.indices_ = id = memory.cache(_build_indices)(X, self._flann_args())
        if get_rhos:
            self.rhos_ = _get_rhos(X, id, Ks, max_K, save_all_Ks, self.min_dist)
        elif hasattr(self, 'rhos_'):
            del self.rhos_
        return self

    def transform(self, X):
        r'''
        Computes the divergences from X to :attr:`features_`.

        Parameters
        ----------
        X : list of bag feature arrays or :class:`skl_groups.features.Features`
            The bags to search "from".

        Returns
        -------
        divs : array of shape ``[len(div_funcs), len(Ks), len(X), len(features_)] + ([2] if do_sym else [])``
            The divergences from X to :attr:`features_`.
            ``divs[d, k, i, j]`` is the ``div_funcs[d]`` divergence
            from ``X[i]`` to ``fetaures_[j]`` using a K of ``Ks[k]``.
            If ``do_sym``, ``divs[d, k, i, j, 0]`` is
            :math:`D_{d,k}( X_i \| \texttt{features_}_j)` and
            ``divs[d, k, i, j, 1]`` is :math:`D_{d,k}(\texttt{features_}_j \| X_i)`.
        '''
        X = as_features(X, stack=True, bare=True)
        Y = self.features_

        Ks = np.asarray(self.Ks)

        if X.dim != Y.dim:
            msg = "incompatible dimensions: fit with {}, transform with {}"
            raise ValueError(msg.format(Y.dim, X.dim))

        memory = self.memory
        if isinstance(memory, string_types):
            memory = Memory(cachedir=memory, verbose=0)

        # ignore Y_indices to avoid slow pickling of them
        # NOTE: if the indices are approximate, then might not get the same
        #       results!
        est = memory.cache(_est_divs, ignore=['n_jobs', 'Y_indices', 'Y_rhos'])
        output, self.rhos_ = est(
            X, Y, self.indices_, getattr(self, 'rhos_', None),
            self.div_funcs, Ks,
            self.do_sym, self.clamp, self.version, self.min_dist,
            self._flann_args(), self._n_jobs)
        return output


def _choose_funcs(div_funcs, Ks, dim, X_n_pts, Y_n_pts, version):
    funcs_base, metas_base, n_meta_only = _parse_specs(div_funcs, Ks)
    funcs, metas = _set_up_funcs(
        funcs_base, metas_base, Ks, dim, X_n_pts, Y_n_pts)

    max_K = Ks.max()
    for func in funcs:
        if hasattr(func, 'K_needed'):
            max_K = max(max_K, func.K_needed)

    save_all_Ks = any(getattr(f, 'needs_all_ks', False) for f in funcs_base)

    # choose the version we can handle
    non_fastable_funcs = set(funcs_base) - fast_funcs
    if version == 'fast':
        if not have_fast_version:
            msg = "asked for 'fast', but skl_groups_accel not available"
            raise ValueError(msg)
        elif non_fastable_funcs:
            msg = "asked for 'fast', but functions are incompatible: {}"
            raise ValueError(msg.format(non_fastable_funcs))
    elif version == 'slow':
        pass
    elif version == 'best':
        if non_fastable_funcs:
            version = 'slow'
        elif have_fast_version:
            version = 'fast'
        else:
            warnings.warn("Using 'slow' version of KNNDivergenceEstimator, "
                          " because skl_groups_accel isn't available; its "
                          "'fast' version is much faster on large "
                          "problems. Pass version='slow' to suppress this "
                          "warning. \n  {}".format(fast_version_error))
            version = 'slow'
    else:
        msg = "Unknown value '{}' for version."
        raise ValueError(msg.format(version))

    return funcs, metas, n_meta_only, max_K, save_all_Ks, version


def _build_indices(X, flann_args):
    "Builds FLANN indices for each bag."
    # TODO: should probably multithread this
    logger.info("Building indices...")
    indices = [None] * len(X)
    for i, bag in enumerate(plog(X, name="index building")):
        indices[i] = idx = FLANNIndex(**flann_args)
        idx.build_index(bag)
    return indices


def _get_rhos(X, indices, Ks, max_K, save_all_Ks, min_dist):
    "Gets within-bag distances for each bag."
    logger.info("Getting within-bag distances...")

    if max_K >= X.n_pts.min():
        msg = "asked for K = {}, but there's a bag with only {} points"
        raise ValueError(msg.format(max_K, X.n_pts.min()))

    # need to throw away the closest neighbor, which will always be self
    # thus K=1 corresponds to column 1 in the result array
    which_Ks = slice(1, None) if save_all_Ks else Ks

    indices = plog(indices, name="within-bag distances")
    rhos = [None] * len(X)
    for i, (idx, bag) in enumerate(zip(indices, X)):
        r = np.sqrt(idx.nn_index(bag, max_K + 1)[1][:, which_Ks])
        np.maximum(min_dist, r, out=r)
        rhos[i] = r
    return rhos


def _est_divs(X, Y, Y_indices, Y_rhos, div_funcs, Ks,
              do_sym, clamp, version, min_dist, flann_args, n_jobs):

    funcs, metas, n_meta_only, max_K, save_all_Ks, version = _choose_funcs(
        div_funcs, Ks, X.dim, X.n_pts, Y.n_pts, version)

    if not do_sym:
        do_sym = {
            req_pos for f, info in iteritems(metas)
                    for req_pos, req in zip(info.deps, f.needs_results)
                    if req.needs_transpose
        }

    to_self = X == Y
    need_Y_rhos = to_self or do_sym or any(meta.needs_rhos[1] for meta in metas)

    if need_Y_rhos:
        if Y_rhos is not None:
            if save_all_Ks:
                old_max_K = Y_rhos[0].shape[1]
                if max_K > old_max_K:
                    msg = ("Y_rhos had a lower max_K ({}) than we actually "
                           "need ({}); recomputing rhos.")
                    logger.warning(msg.format(old_max_K, max_K))
                    Y_rhos = None
            else:
                if Y_rhos[0].shape[1] != Ks.size:
                    raise ValueError("invalid Y_rhos passed")

        if Y_rhos is None:
            Y_rhos = _get_rhos(Y, Y_indices, Ks, max_K, save_all_Ks, min_dist)

    if to_self:
        X_indices = Y_indices
        X_rhos = Y_rhos
    else:
        X_indices = _build_indices(X, flann_args)
        X_rhos = _get_rhos(X, X_indices, Ks, max_K, save_all_Ks, min_dist)

    logger.info("Getting divergences...")
    if version == 'fast':
        fn = _estimate_cross_divs_fast
    else:
        fn = _estimate_cross_divs
    outputs = fn(X, X_indices, X_rhos, Y, Y_indices, Y_rhos,
                 funcs, Ks, max_K, save_all_Ks, len(div_funcs) + n_meta_only,
                 do_sym, to_self,
                 ProgressLogger(progress_logger, name="cross-divergences"),
                 n_jobs, min_dist, clamp)

    logger.info("Computing meta-divergences...")

    if save_all_Ks:
        X_rhos = [rho[:, Ks - 1] for rho in X_rhos]
        if need_Y_rhos:
            Y_rhos = [rho[:, Ks - 1] for rho in Y_rhos]

    for meta, info in iteritems(metas):
        args = ()
        if meta.needs_rhos[0]:
            args += (X_rhos,)
        if meta.needs_rhos[1]:
            args += (Y_rhos,)
        args += (outputs[info.deps],)

        outputs[info.pos] = meta(*args, clamp=clamp, to_self=to_self)

    if do_sym != True:
        outputs = outputs[:, :, :, :, 0]

    if n_meta_only:
        outputs = np.ascontiguousarray(outputs[:-n_meta_only])

    logger.info("Done with divergences.")
    return outputs, Y_rhos


################################################################################
### Estimators of various divergences based on nearest-neighbor distances.
#
# The standard interface for these functions is:
#
# Function attributes:
#
#   needs_alpha: whether this function needs an alpha parameter. Default false.
#
#   self_value: The value that this function should take when comparing a
#               sample to itself: either a scalar constant or None (the
#               default), in which case the function is still called with
#               rhos = nus.
#
#   chooser_fn: a function that gets alphas (if needs_alpha), Ks, dim, X_ns, Y_ns
#               (the arrays of bag sizes) and returns a partial() of a "core"
#               function, with some things precomputed. If not present, just
#               does partial(fn, [alphas,] Ks, dim).
#
#   needs_all_ks: whether this function needs *all* the neighbor distances up
#                 to the max K value, rather than just the values of K that are
#                 actually used. Default false.
#
#   chooser_fn.returns_ks: whether the chooser_fn returns the max value of K
#                          needed. This allows an estimator function to require
#                          a higher value of K than requested by the user. Only
#                          if needs_all_ks; default false.
#
# Arguments:
#
#   alphas (if needs_alpha; array-like, scalar or 1d): the alpha values to use
#
#   Ks (array-like, scalar or 1d): the K values used
#
#   dim (scalar): the dimension of the feature space
#
#   num_q (scalar): the number of points in the sample from q
#
#   rhos: an array of within-bag nearest neighbor distances for a sample from p.
#         rhos[i, j] should be the distance from the ith sample from p to its
#         Ks[j]'th neighbor in the same sample. Shape: (num_p, num_Ks).
#   nus: an array of nearest neighbor distances from samples from other dists.
#        nus[i, j] should be the distance from the ith sample from p to its
#        Ks[j]'th neighbor in the sample from q. Shape: (num_p, num_Ks).
#
# Returns an array of divergence estimates. If needs_alpha, should be of shape
# (num_alphas, num_Ks); otherwise, of shape (num_Ks,).

def linear(Ks, dim, num_q, rhos, nus):
    r'''
    Estimates the linear inner product \int p q between two distributions,
    based on kNN distances.
    '''
    return _get_linear(Ks, dim)(num_q, rhos, nus)

def _get_linear(Ks, dim, X_ns=None, Y_ns=None):
    # Estimated with alpha=0, beta=1:
    #   B_{k,d,0,1} = (k - 1) / pi^(dim/2) * gamma(dim/2 + 1)
    #   (using gamma(k) / gamma(k - 1) = k - 1)
    Ks = np.reshape(Ks, (-1,))
    Bs = (Ks - 1) / np.pi ** (dim / 2) * gamma(dim / 2 + 1)  # shape (num_Ks,)
    return partial(_linear, Bs, dim)
linear.self_value = None  # have to execute it
linear.needs_alpha = False
linear.chooser_fn = _get_linear

# kl function is entirely in _np_divs (nothing to precompute)

def alpha_div(alphas, Ks, dim, num_q, rhos, nus):
    r'''
    Estimate the alpha divergence between distributions:
        \int p^\alpha q^(1-\alpha)
    based on kNN distances.

    Used in Renyi, Hellinger, Bhattacharyya, Tsallis divergences.

    Enforces that estimates are >= 0.

    Returns divergence estimates with shape (num_alphas, num_Ks).
    '''
    return _get_alpha_div(alphas, Ks, dim)(num_q, rhos, nus)

def _get_alpha_div(alphas, Ks, dim, X_ns=None, Y_ns=None):
    alphas = np.reshape(alphas, (-1, 1))
    Ks = np.reshape(Ks, (1, -1))

    omas = 1 - alphas

    # We're estimating with alpha = alpha-1, beta = 1-alpha.
    # B constant in front:
    #   estimator's alpha = -beta, so volume of unit ball cancels out
    #   and then ratio of gamma functions
    Bs = np.exp(gammaln(Ks) * 2 - gammaln(Ks + omas) - gammaln(Ks - omas))

    return partial(_alpha_div, omas, Bs, dim)

alpha_div.self_value = 1
alpha_div.needs_alpha = True
alpha_div.chooser_fn = _get_alpha_div


def jensen_shannon_core(Ks, dim, num_q, rhos, nus):
    r'''
    Estimates
          1/2 mean_X( d * log radius of largest ball in X+Y around X_i
                                with no more than M/(n+m-1) weight
                                where X points have weight 1 / (2 n - 1)
                                  and Y points have weight n / (m (2 n - 1))
                      - digamma(# of neighbors in that ball))

    This is the core pairwise component of the estimator of Jensen-Shannon
    divergence based on the Hino-Murata weighted information estimator. See
    the docstring for jensen_shannon for an explanation.
    '''
    ns = np.array([rhos.shape[0], num_q])
    return _get_jensen_shannon_core(Ks, dim, ns)[0](num_q, rhos, nus)

def _get_jensen_shannon_core(Ks, dim, X_ns, Y_ns):
    # precompute the max/min possible digamma(i) values: the floors/ceils of
    #
    #   M/(n+m-1) / (1 / (2 n - 1))
    #   M/(n+m-1) / (n / (m (2 n - 1)))
    #
    # for any valid value of n, m.

    min_X_n = np.min(X_ns)
    max_X_n = np.max(X_ns)
    if Y_ns is None:
        min_Y_n = min_X_n
        max_Y_n = max_X_n
    else:
        min_Y_n = np.min(Y_ns)
        max_Y_n = np.max(Y_ns)
    min_K = np.min(Ks)
    max_K = np.max(Ks)

    # figure out the smallest i value we might need (# of neighbors in ball)

    wt_bounds = [np.inf, -np.inf]
    min_wt_n = None; min_wt_m = None
    # max_wt_n = None; max_wt_m = None
    n_ms = list(itertools.product([min_X_n, max_X_n], [min_Y_n, max_Y_n]))
    for n, m in itertools.chain(n_ms, map(reversed, n_ms)):
        base = (2 * n - 1) / (n + m - 1)

        for wt in (base, base * m / n):
            if wt < wt_bounds[0]:
                wt_bounds[0] = wt
                min_wt_n = n
                min_wt_m = m

            if wt > wt_bounds[1]:
                wt_bounds[1] = wt
                # max_wt_n = n
                # max_wt_m = m

    if wt_bounds[0] * min_K < 1:
        msg = "K={} is too small for Jensen-Shannon estimator with n={}, m={}"
        raise ValueError((msg + "; must be at least {}").format(
             min_K, min_wt_n, min_wt_m, int(np.ceil(1 / wt_bounds[0]))))

    min_i = int(np.floor(wt_bounds[0] * min_K))
    max_i = int(np.ceil( wt_bounds[1] * max_K))
    digamma_vals = psi(np.arange(min_i, max_i + 1))

    # TODO: If we don't actually hit the worst case, might be nice to still
    #       run and just nan those elements that we can't compute. This is
    #       over-conservative.
    return partial(_jensen_shannon_core, Ks, dim, min_i, digamma_vals), max_i

jensen_shannon_core.needs_alpha = False
jensen_shannon_core.chooser_fn = _get_jensen_shannon_core
jensen_shannon_core.needs_all_ks = True
jensen_shannon_core.chooser_fn.returns_ks = True
jensen_shannon_core.self_value = np.nan
# The self_value should be the entropy estimate. But since we'll just subtract
# that later, don't bother computing it.


################################################################################
### Meta-estimators: things that need some additional computation on top of
###                  the per-bag stuff of the functions above.

# These functions are run after the base estimators above are complete.
#
# The interface here is:
#
# Function attributes:
#
#   needs_alpha: whether this function needs an alpha parameter. Default false.
#
#   needs_results: a list of MetaRequirement objects (below).
#                  Note that it is legal for meta estimators to depend on other
#                  meta estimators; circular dependencies cause the spec parser
#                  to crash.
#
# Arguments:
#
#   alphas (if needs_alpha; array-like, scalar or 1d): the alpha values to use
#
#   Ks (array-like, scalar or 1d): the K values used
#
#   dim (scalar): the dimension of the feature space
#
#   rhos: a list of within-bag NN distances, each of which is like the rhos
#         argument above
#
#   required: a list of the results array for each MetaRequirement classes,
#             each of shape (num_Ks, n_X, n_Y, 1 or 2),
#             where the last dimension depends on whether we're doing the
#             symmetric or not.
#
# Returns: array of results, shape (num_alphas, num_Ks, n_X, n_Y, 1 or 2).

MetaRequirement = namedtuple('MetaRequirement', 'func alpha needs_transpose')
# func: the function of the regular divergence that's needed
# alpha: None if no alpha is needed. Otherwise, can be a scalar alpha value,
#        or a callable which takes the (scalar or list) alphas for the meta
#        function and returns the required function's alpha(s).
# needs_transpose: if true, ensures the required results have both directions


def bhattacharyya(Ks, dim, required, clamp=True, to_self=False):
    r'''
    Estimate the Bhattacharyya coefficient between distributions, based on kNN
    distances:  \int \sqrt{p q}

    If clamp (the default), enforces 0 <= BC <= 1.

    Returns an array of shape (num_Ks,).
    '''
    est = required
    if clamp:
        est = np.minimum(est, 1)  # BC <= 1
    return est
bhattacharyya.needs_alpha = False
bhattacharyya.needs_rhos = (False, False)
bhattacharyya.needs_results = [MetaRequirement(alpha_div, 0.5, False)]


def hellinger(Ks, dim, required, clamp=True, to_self=False):
    r'''
    Estimate the Hellinger distance between distributions, based on kNN
    distances:  \sqrt{1 - \int \sqrt{p q}}

    Always enforces 0 <= H, to be able to sqrt; if clamp, also enforces
    H <= 1.

    Returns a vector: one element for each K.
    '''
    bc = required
    est = 1 - bc
    np.maximum(est, 0, out=est)
    if clamp:
        np.minimum(est, 1, out=est)
    np.sqrt(est, out=est)
    return est
hellinger.needs_alpha = False
hellinger.needs_rhos = (False, False)
hellinger.needs_results = [MetaRequirement(alpha_div, 0.5, False)]


def renyi(alphas, Ks, dim, required, min_val=np.spacing(1),
          clamp=True, to_self=False):
    r'''
    Estimate the Renyi-alpha divergence between distributions, based on kNN
    distances:  1/(\alpha-1) \log \int p^alpha q^(1-\alpha)

    If the inner integral is less than min_val (default ``np.spacing(1)``),
    uses the log of min_val instead.

    If clamp (the default), enforces that the estimates are nonnegative by
    replacing any negative estimates with 0.

    Returns an array of shape (num_alphas, num_Ks).
    '''
    alphas = np.reshape(alphas, (-1, 1))
    est = required

    est = np.maximum(est, min_val)  # TODO: can we modify in-place?
    np.log(est, out=est)
    est /= alphas - 1
    if clamp:
        np.maximum(est, 0, out=est)
    return est
renyi.needs_alpha = True
renyi.needs_rhos = (False, False)
renyi.needs_results = [MetaRequirement(alpha_div, identity, False)]


def tsallis(alphas, Ks, dim, required, clamp=True, to_self=False):
    r'''
    Estimate the Tsallis-alpha divergence between distributions, based on kNN
    distances:  (\int p^alpha q^(1-\alpha) - 1) / (\alpha - 1)

    If clamp (the default), enforces the estimate is nonnegative.

    Returns an array of shape (num_alphas, num_Ks).
    '''
    alphas = np.reshape(alphas, (-1, 1))
    alpha_est = required

    est = alpha_est - 1
    est /= alphas - 1
    if clamp:
        np.maximum(est, 0, out=est)
    return est
tsallis.needs_alpha = True
tsallis.needs_rhos = (False, False)
tsallis.needs_results = [MetaRequirement(alpha_div, identity, False)]


def l2(Ks, dim, X_rhos, Y_rhos, required, clamp=True, to_self=False):
    r'''
    Estimates the L2 distance between distributions, via
        \int (p - q)^2 = \int p^2 - \int p q - \int q p + \int q^2.

    \int pq and \int qp are estimated with the linear function (in both
    directions), while \int p^2 and \int q^2 are estimated via the quadratic
    function below.

    Always clamps negative estimates of l2^2 to 0, because otherwise the sqrt
    would break.
    '''
    n_X = len(X_rhos)
    n_Y = len(Y_rhos)

    linears = required
    assert linears.shape == (1, Ks.size, n_X, n_Y, 2)

    X_quadratics = np.empty((Ks.size, n_X), dtype=np.float32)
    for i, rho in enumerate(X_rhos):
        X_quadratics[:, i] = quadratic(Ks, dim, rho)

    Y_quadratics = np.empty((Ks.size, n_Y), dtype=np.float32)
    for j, rho in enumerate(Y_rhos):
        Y_quadratics[:, j] = quadratic(Ks, dim, rho)

    est = -linears.sum(axis=4)
    est += X_quadratics[None, :, :, None]
    est += Y_quadratics[None, :, None, :]
    np.maximum(est, 0, out=est)
    np.sqrt(est, out=est)

    # diagonal is of course known to be zero
    if to_self:
        est[:, :, xrange(n_X), xrange(n_Y)] = 0
    return est[:, :, :, :, None]
l2.needs_alpha = False
l2.needs_rhos = (True, True)
l2.needs_results = [MetaRequirement(linear, alpha=None, needs_transpose=True)]


# Not actually a meta-estimator, though it could be if it just repeated the
# values across rows (or columns).
def quadratic(Ks, dim, rhos, required=None):
    r'''
    Estimates \int p^2 based on kNN distances.

    In here because it's used in the l2 distance, above.

    Returns array of shape (num_Ks,).
    '''
    # Estimated with alpha=1, beta=0:
    #   B_{k,d,1,0} is the same as B_{k,d,0,1} in linear()
    # and the full estimator is
    #   B / (n - 1) * mean(rho ^ -dim)
    N = rhos.shape[0]
    Ks = np.asarray(Ks)
    Bs = (Ks - 1) / np.pi ** (dim / 2) * gamma(dim / 2 + 1)  # shape (num_Ks,)
    est = Bs / (N - 1) * np.mean(rhos ** (-dim), axis=0)
    return est


def jensen_shannon(Ks, dim, X_rhos, Y_rhos, required,
                   clamp=True, to_self=False):
    r'''
    Estimate the difference between the Shannon entropy of an equally-weighted
    mixture between X and Y and the mixture of the Shannon entropies:

        JS(X, Y) = H[ (X + Y) / 2 ] - (H[X] + H[Y]) / 2

    We use a special case of the Hino-Murata weighted information estimator with
    a fixed M = n \alpha, about equivalent to the K-nearest-neighbor approach
    used for the other estimators:

        Hideitsu Hino and Noboru Murata (2013).
        Information estimators for weighted observations. Neural Networks.
        http://linkinghub.elsevier.com/retrieve/pii/S0893608013001676


    The estimator for JS(X, Y) is:

        log volume of the unit ball - log M + log(n + m - 1) + digamma(M)
        + 1/2 mean_X( d * log radius of largest ball in X+Y around X_i
                                with no more than M/(n+m-1) weight
                                where X points have weight 1 / (2 n - 1)
                                  and Y points have weight n / (m (2 n - 1))
                      - digamma(# of neighbors in that ball) )
        + 1/2 mean_Y( d * log radius of largest ball in X+Y around Y_i
                                with no more than M/(n+m-1) weight
                                where X points have weight m / (n (2 m - 1))
                                  and Y points have weight 1 / (2 m - 1)
                      - digamma(# of neighbors in that ball) )

        - 1/2 (log volume of the unit ball - log M + log(n - 1) + digamma(M))
        - 1/2 mean_X( d * log radius of the largest ball in X around X_i
                                with no more than M/(n-1) weight
                                where X points have weight 1 / (n - 1))
                      - digamma(# of neighbors in that ball) )

        - 1/2 (log volume of the unit ball - log M + log(m - 1) + digamma(M))
        - 1/2 mean_Y( d * log radius of the largest ball in Y around Y_i
                                with no more than M/(n-1) weight
                                where X points have weight 1 / (m - 1))
                      - digamma(# of neighbors in that ball) )

        =

        log(n + m - 1) + digamma(M)
        + 1/2 mean_X( d * log radius of largest ball in X+Y around X_i
                                with no more than M/(n+m-1) weight
                                where X points have weight 1 / (2 n - 1)
                                  and Y points have weight n / (m (2 n - 1))
                      - digamma(# of neighbors in that ball) )
        + 1/2 mean_Y( d * log radius of largest ball in X+Y around Y_i
                                with no more than M/(n+m-1) weight
                                where X points have weight m / (n (2 m - 1))
                                  and Y points have weight 1 / (2 m - 1)
                      - digamma(# of neighbors in that ball) )
        - 1/2 [log(n-1) + mean_X( d * log rho_M(X_i) )]
        - 1/2 [log(m-1) + mean_Y( d * log rho_M(Y_i) )]
    '''

    X_ns = np.array([rho.shape[0] for rho in X_rhos])
    Y_ns = np.array([rho.shape[0] for rho in Y_rhos])
    n_X = X_ns.size
    n_Y = Y_ns.size

    # cores[0, k, i, j, 0] is mean_X(d * ... - psi(...)) for X[i], Y[j], M=Ks[k]
    # cores[0, k, i, j, 1] is mean_Y(d * ... - psi(...)) for X[i], Y[j], M=Ks[k]
    cores = required
    assert cores.shape == (1, Ks.size, n_X, n_Y, 2)

    # X_bits[k, i] is log(n-1) + mean_X( d * log rho_M(X_i) )  for X[i], M=Ks[k]
    X_bits = np.empty((Ks.size, n_X), dtype=np.float32)
    for i, rho in enumerate(X_rhos):
        X_bits[:, i] = dim * np.mean(np.log(rho), axis=0)
    X_bits += np.log(X_ns - 1)[np.newaxis, :]

    # Y_bits[k, j] is log(n-1) + mean_Y( d * log rho_M(Y_i) )  for Y[j], M=Ks[k]
    Y_bits = np.empty((Ks.size, n_Y), dtype=np.float32)
    for j, rho in enumerate(Y_rhos):
        Y_bits[:, j] = dim * np.mean(np.log(rho), axis=0)
    Y_bits += np.log(Y_ns - 1)[np.newaxis, :]

    est = cores.sum(axis=4)
    est -= X_bits.reshape(1, Ks.size, n_X, 1)
    est -= Y_bits.reshape(1, Ks.size, 1, n_Y)
    est /= 2
    est += np.log(-1 + X_ns[None, None, :, None] + Y_ns[None, None, None, :])
    est += psi(Ks)[None, :, None, None]

    # diagonal is zero
    if to_self:
        est[:, :, xrange(n_X), xrange(n_Y)] = 0

    if clamp:  # know that 0 <= JS <= ln(2)
        np.maximum(0, est, out=est)
        np.minimum(np.log(2), est, out=est)
    return est[:, :, :, :, None]
jensen_shannon.needs_alpha = False
jensen_shannon.needs_rhos = (True, True)
jensen_shannon.needs_results = [
    MetaRequirement(jensen_shannon_core, alpha=None, needs_transpose=True)]


################################################################################
### Parse string specifications into functions to use

func_mapping = {
    'linear': linear,
    'kl': kl,
    'alpha': alpha_div,
    'bc': bhattacharyya,
    'hellinger': hellinger,
    'renyi': renyi,
    'tsallis': tsallis,
    'l2': l2,
    'js-core': jensen_shannon_core,
    'js': jensen_shannon,
    'jensen-shannon': jensen_shannon,
}
fast_funcs = {kl, alpha_div, linear, jensen_shannon_core}


def topological_sort(deps):
    '''
    Topologically sort a DAG, represented by a dict of child => set of parents.
    The dependency dict is destroyed during operation.

    Uses the Kahn algorithm: http://en.wikipedia.org/wiki/Topological_sorting
    Not a particularly good implementation, but we're just running it on tiny
    graphs.
    '''
    order = []
    available = set()

    def _move_available():
        to_delete = []
        for n, parents in iteritems(deps):
            if not parents:
                available.add(n)
                to_delete.append(n)
        for n in to_delete:
            del deps[n]

    _move_available()
    while available:
        n = available.pop()
        order.append(n)
        for parents in itervalues(deps):
            parents.discard(n)
        _move_available()

    if available:
        raise ValueError("dependency cycle found")
    return order


_FuncInfo = namedtuple('_FuncInfo', 'alphas pos')
_MetaFuncInfo = namedtuple('_MetaFuncInfo', 'alphas pos deps')
def _parse_specs(specs, Ks):
    '''
    Set up the different functions we need to call.

    Returns:
        - a dict mapping base estimator functions to _FuncInfo objects.
          If the function needs_alpha, then the alphas attribute is an array
          of alpha values and pos is a corresponding array of indices.
          Otherwise, alphas is None and pos is a list containing a single index.
          Indices are >= 0 if they correspond to something in a spec,
          and negative if they're just used for a meta estimator but not
          directly requested.
        - an OrderedDict mapping functions to _MetaFuncInfo objects.
          alphas and pos are like for _FuncInfo; deps is a list of indices
          which should be passed to the estimator. Note that these might be
          other meta functions; this list is guaranteed to be in an order
          such that all dependencies are resolved before calling that function.
          If no such order is possible, raise ValueError.
        - the number of meta-only results

    # TODO: update doctests for _parse_specs

    >>> _parse_specs(['renyi:.8', 'hellinger', 'renyi:.9'])
    ({<function alpha_div at 0x10954f848>:
            _FuncInfo(alphas=[0.8, 0.5, 0.9], pos=[-1, -2, -3])},
     OrderedDict([
        (<function hellinger at 0x10954fc80>,
            _MetaFuncInfo(alphas=None, pos=[1], deps=[array(-2)])),
        (<function renyi at 0x10954fcf8>,
            _MetaFuncInfo(alphas=[0.8, 0.9], pos=[0, 2], deps=[-1, -3]))
     ]), 3)

    >>> _parse_specs(['renyi:.8', 'hellinger', 'renyi:.9', 'l2'])
    ({<function alpha_div at 0x10954f848>:
        _FuncInfo(alphas=[0.8, 0.5, 0.9], pos=[-1, -2, -3]),
      <function linear at 0x10954f758>: _FuncInfo(alphas=None, pos=[-4])
     }, OrderedDict([
        (<function hellinger at 0x10954fc80>,
            _MetaFuncInfo(alphas=None, pos=[1], deps=[array(-2)])),
        (<function l2 at 0x10954fde8>,
            _MetaFuncInfo(alphas=None, pos=[3], deps=[-4])),
        (<function renyi at 0x10954fcf8>,
            _MetaFuncInfo(alphas=[0.8, 0.9], pos=[0, 2], deps=[-1, -3]))
     ]), 4)

    >>> _parse_specs(['renyi:.8', 'hellinger', 'renyi:.9', 'l2', 'linear'])
    ({<function alpha_div at 0x10954f848>:
        _FuncInfo(alphas=[0.8, 0.5, 0.9], pos=[-1, -2, -3]),
      <function linear at 0x10954f758>: _FuncInfo(alphas=None, pos=[4])
     }, OrderedDict([
        (<function hellinger at 0x10954fc80>,
            _MetaFuncInfo(alphas=None, pos=[1], deps=[array(-2)])),
        (<function l2 at 0x10954fde8>,
            _MetaFuncInfo(alphas=None, pos=[3], deps=[4])),
        (<function renyi at 0x10954fcf8>,
            _MetaFuncInfo(alphas=[0.8, 0.9], pos=[0, 2], deps=[-1, -3]))
     ]), 3)
    '''
    funcs = {}
    metas = {}
    meta_deps = defaultdict(set)

    def add_func(func, alpha=None, pos=None):
        needs_alpha = getattr(func, 'needs_alpha', False)
        is_meta = hasattr(func, 'needs_results')

        d = metas if is_meta else funcs
        if func not in d:
            if needs_alpha:
                args = {'alphas': [alpha], 'pos': [pos]}
            else:
                args = {'alphas': None, 'pos': [pos]}

            if not is_meta:
                d[func] = _FuncInfo(**args)
            else:
                d[func] = _MetaFuncInfo(deps=[], **args)
                for req in func.needs_results:
                    if callable(req.alpha):
                        req_alpha = req.alpha(alpha)
                    else:
                        req_alpha = req.alpha
                    add_func(req.func, alpha=req_alpha)
                    meta_deps[func].add(req.func)
                    meta_deps[req.func]  # make sure required func is in there

        else:
            # already have an entry for the func
            # need to give it this pos, if it's not None
            # and also make sure that the alpha is present
            info = d[func]
            if not needs_alpha:
                if pos is not None:
                    if info.pos != [None]:
                        msg = "{} passed more than once"
                        raise ValueError(msg.format(func_name))

                    info.pos[0] = pos
            else:  # needs alpha
                try:
                    idx = info.alphas.index(alpha)
                except ValueError:
                    # this is a new alpha value we haven't seen yet
                    info.alphas.append(alpha)
                    info.pos.append(pos)
                    if is_meta:
                        for req in func.needs_results:
                            if callable(req.alpha):
                                req_alpha = req.alpha(alpha)
                            else:
                                req_alpha = req.alpha
                            add_func(req.func, alpha=req_alpha)
                else:
                    # repeated alpha value
                    if pos is not None:
                        if info.pos[idx] is not None:
                            msg = "{} with alpha {} passed more than once"
                            raise ValueError(msg.format(func_name, alpha))
                        info.pos[idx] = pos

    # add functions for each spec
    for i, spec in enumerate(specs):
        func_name, alpha = (spec.split(':', 1) + [None])[:2]
        if alpha is not None:
            alpha = float(alpha)

        try:
            func = func_mapping[func_name]
        except KeyError:
            msg = "'{}' is not a known function type"
            raise ValueError(msg.format(func_name))

        needs_alpha = getattr(func, 'needs_alpha', False)
        if needs_alpha and alpha is None:
            msg = "{} needs alpha but not passed in spec '{}'"
            raise ValueError(msg.format(func_name, spec))
        elif not needs_alpha and alpha is not None:
            msg = "{} doesn't need alpha but is passed in spec '{}'"
            raise ValueError(msg.format(func_name, spec))

        add_func(func, alpha, i)

    # number things that are dependencies only
    meta_counter = itertools.count(-1, step=-1)
    for info in itertools.chain(itervalues(funcs), itervalues(metas)):
        for i, pos in enumerate(info.pos):
            if pos is None:
                info.pos[i] = next(meta_counter)

    # fill in the dependencies for metas
    for func, info in iteritems(metas):
        deps = info.deps
        assert deps == []

        for req in func.needs_results:
            f = req.func
            req_info = (metas if hasattr(f, 'needs_results') else funcs)[f]
            if req.alpha is not None:
                if callable(req.alpha):
                    req_alpha = req.alpha(info.alphas)
                else:
                    req_alpha = req.alpha

                find_alpha = np.vectorize(req_info.alphas.index, otypes=[int])
                pos = np.asarray(req_info.pos)[find_alpha(req_alpha)]
                if np.isscalar(pos):
                    deps.append(pos[()])
                else:
                    deps.extend(pos)
            else:
                pos, = req_info.pos
                deps.append(pos)

    # topological sort of metas
    meta_order = topological_sort(meta_deps)
    metas_ordered = OrderedDict(
        (f, metas[f]) for f in meta_order if hasattr(f, 'needs_results'))

    return funcs, metas_ordered, -next(meta_counter) - 1


def _set_up_funcs(funcs, metas_ordered, Ks, dim, X_ns=None, Y_ns=None):
    # replace functions with partials of args
    def replace_func(func, info):
        needs_alpha = getattr(func, 'needs_alpha', False)

        new = None
        args = (Ks, dim)
        if needs_alpha:
            args = (info.alphas,) + args

        if hasattr(func, 'chooser_fn'):
            args += (X_ns, Y_ns)
            if (getattr(func, 'needs_all_ks', False) and
                    getattr(func.chooser_fn, 'returns_ks', False)):
                new, K = func.chooser_fn(*args)
                new.K_needed = K
            else:
                new = func.chooser_fn(*args)
        else:
            new = partial(func, *args)

        for attr in dir(func):
            if not (attr.startswith('__') or attr.startswith('func_')):
                setattr(new, attr, getattr(func, attr))
        return new

    rep_funcs = dict(
        (replace_func(f, info), info) for f, info in iteritems(funcs))
    rep_metas_ordered = OrderedDict(
        (replace_func(f, info), info) for f, info in iteritems(metas_ordered))

    return rep_funcs, rep_metas_ordered



from __future__ import division

import operator as op

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.externals.six.moves import xrange, reduce

from ..features import as_features
from ..utils import is_integer

VALID_BASES = frozenset({'cosine'})
class L2DensityTransformer(BaseEstimator, TransformerMixin):
    '''
    Summarizes sets into a single vector, whose inner products estimate the
    L2 inner product between the density functions underlying each of the sets.

    **IMPORTANT**: Assumes that inputs lie in [0, 1] - use
    :class:`skl_groups.preprocessing.BagMinMaxScaler`
    with ``feature_range=(0, 1)``, ``truncate=True``, and optionally a
    ``fit_feature_range`` to ensure this.

    Classifiers built off of this representation can scale to very large set
    sizes (since they're summarized in a fixed-dimensional vector) and/or
    number of sets (since you can do a linear classifier on either them or
    an explicit kernel mapping). But the representation only works for very
    small number of dimensions: the dimensionality of the output is exponential
    in both the smoothness coefficient and the dimension of the input. Some
    specific numbers:

    ==========  =======  =======  =======  =======  =======
    smoothness      d=1      d=2      d=3      d=4      d=5
    ==========  =======  =======  =======  =======  =======
             1        2        3        4        5        6
             2        3        6       11       20       36
             3        4       11       29       70      157
             4        5       17       54      165      482
             5        6       26       99      357    1,203
             6        7       35      163      688    2,673
             7        8       45      239    1,154    5,139
             8        9       58      344    1,867    9,389
             9       10       73      486    2,907   15,930
            10       11       90      648    4,272   25,810
            11       12      106      847    6,070   39,855
            12       13      123    1,069    8,357   59,711
            13       14      146    1,355   11,307   86,346
            14       15      168    1,680   14,998  122,467
            15       16      193    2,046   19,470  168,957
            16       17      216    2,446   24,809  229,743
            17       18      243    2,911   31,211  305,153
            18       19      271    3,443   38,928  400,844
            19       20      302    4,022   47,816  518,154
            20       21      335    4,662   58,201  662,629
    ==========  =======  =======  =======  =======  =======

    For more on the estimator, see [1]_, which effectively uses a pipeline of
    :class:`skl_groups.preprocessing.BagMinMaxScaler`,
    :class:`L2DensityTransformer`,
    :class:`sklearn.kernel_approximation.RBFSampler`,
    and then :class:`sklearn.linear_model.Ridge`.

    Parameters
    ----------
    smoothness : positive int
        The smoothness assumption; lower values mean "smoother", with respect
        to the underlying basis. Higher values make the dimensionality higher.

    basis : 'cosine', optional, default 'cosine'
        The underlying orthonormal function basis to use.
        'cosine', the only current option, uses a Fourier basis of
        cosines with differently-scaled periods.

    do_bounds_check : boolean, optional, default True
        Whether to make sure that the data is in [0, 1] in :meth:`transform`.
        (If you skip this check, and the data isn't in the right scale, you'll
        get silently incorrect results.)
    
    Attributes
    ----------
    `inds_` : integer array of shape [output_dim, input_dim]
        The indices of the basis to use. Output dimension i is the
        product of the value of the ``inds_[i, j]``'th basis function at the
        ``i``'th input coordinate.

    References
    ----------
    .. [1] J. B. Oliva, B. Poczos, & J. Schneider (2014).
           Fast Distribution To Real Regression.
           In Proceedings of the 17th International Conference on
           Artificial Intelligence and Statistics (AISTATS).
    '''

    def __init__(self, smoothness, basis='cosine', do_bounds_check=True):
        self.smoothness = smoothness
        if basis not in VALID_BASES:
            raise ValueError("unknown basis '{}'".format(basis))
        self.basis = basis
        self.do_bounds_check = do_bounds_check

    def _check_fitted(self):
        if not hasattr(self, "inds_"):
            raise AttributeError("Model has not been trained yet.")

    @property
    def dim_(self):
        "The dimension of the inputs, once fitted."
        self._check_fitted()
        return self.inds_.shape[1]

    @property
    def out_dim_(self):
        "The dimension of the output vectors, once fitted."
        self._check_fitted()
        return self.inds_.shape[0]

    def fit(self, X, y=None):
        '''
        Picks the elements of the basis to use for the given data.

        Only depends on the dimension of X. If it's more convenient, you can
        pass a single integer for X, which is the dimension to use.

        Parameters
        ----------
        X : an integer, a :class:`Features` instance, or a list of bag features
            The input data, or just its dimension, since only the dimension is
            needed here.
        '''
        if is_integer(X):
            dim = X
        else:
            X = as_features(X)
            dim = X.dim
        M = self.smoothness

        # figure out the smooth-enough elements of our basis
        inds = np.mgrid[(slice(M + 1),) * dim].reshape(dim, (M + 1) ** dim).T
        self.inds_ = inds[(inds ** 2).sum(axis=1) <= M ** 2]
        return self

    def transform(self, X):
        '''
        Transform a list of bag features into its projection series
        representation.

        Parameters
        ----------
        X : :class:`skl_groups.features.Features` or list of bag feature arrays
            New data to transform. The data should all lie in [0, 1];
            use :class:`skl_groups.preprocessing.BagMinMaxScaler` if not.

        Returns
        -------
        X_new : integer array, shape ``[len(X), dim_]``
            X transformed into the new space.
        '''
        self._check_fitted()
        M = self.smoothness
        dim = self.dim_
        inds = self.inds_
        do_check = self.do_bounds_check

        X = as_features(X)
        if X.dim != dim:
            msg = "model fit for dimension {} but got dim {}"
            raise ValueError(msg.format(dim, X.dim))

        Xt = np.empty((len(X), self.inds_.shape[0]))
        Xt.fill(np.nan)

        if self.basis == 'cosine':  # TODO: put this in a C extension?
            coefs = (np.pi * np.arange(M + 1))[..., :]
            for i, bag in enumerate(X):
                if do_check:
                    if np.min(bag) < 0 or np.max(bag) > 1:
                        raise ValueError("Bag {} not in [0, 1]".format(i))

                # apply each phi func to each dataset point: n x dim x M
                phi = coefs * bag[..., np.newaxis]
                np.cos(phi, out=phi)
                phi[:, :, 1:] *= np.sqrt(2)

                # B is the evaluation of each tensor-prodded basis func
                # at each point: n x inds.shape[0]
                B = reduce(op.mul, (phi[:, i, inds[:, i]] for i in xrange(dim)))

                Xt[i, :] = np.mean(B, axis=0)
        else:
            raise ValueError("unknown basis '{}'".format(self.basis))

        return Xt



from __future__ import division

import os
import sys

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.externals.six.moves import xrange
from nose.tools import assert_raises

if __name__ == '__main__':
    # make this copy of skl_groups importable
    _this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(_this_dir)))

from skl_groups.features import Features
from skl_groups.preprocessing import (BagPreprocesser, BagStandardizer,
                                      BagMinMaxScaler, BagNormalizer, BagPCA)


################################################################################

def test_basic():
    bags = [np.random.normal(5, 3, size=(np.random.randint(10, 100), 20))
            for _ in xrange(50)]
    feats = Features(bags, stack=True)

    stder = BagStandardizer()
    stdized = stder.fit_transform(bags)
    stdized.make_stacked()

    assert np.allclose(np.mean(stdized.stacked_features), 0)
    assert np.allclose(np.std(stdized.stacked_features), 1)

    first_five = stder.transform(bags[:5])
    assert first_five == stdized[:5]

    minmaxer = BagMinMaxScaler([3, 7])
    minmaxed = minmaxer.fit_transform(feats)
    minmaxed.make_stacked()
    assert np.allclose(np.min(minmaxed.stacked_features, 0), 3)
    assert np.allclose(np.max(minmaxed.stacked_features, 0), 7)

    normer = BagNormalizer('l1')
    normed = normer.fit_transform(Features(bags))
    normed.make_stacked()
    assert np.allclose(np.sum(np.abs(normed.stacked_features), 1), 1)

    class GetMean(BaseEstimator, TransformerMixin):
        def fit(self, X, y=None):
            return self
        def transform(self, X):
            return X.mean(axis=1)[None, :]
    m = BagPreprocesser(GetMean())
    assert_raises(ValueError, lambda: m.transform(bags))


def test_pca():
    bags = [np.random.normal(5, 3, size=(np.random.randint(10, 100), 20))
            for _ in xrange(50)]
    feats = Features(bags, stack=True)

    pca = BagPCA(k=3)
    pca.fit(bags)
    pcaed = pca.transform(bags)
    assert pcaed.dim == 3

    BagPCA(varfrac=.3).fit_transform(bags)

    pca2 = BagPCA(k=20)
    pcaed2 = pca2.fit_transform(bags)
    orig = pca2.inverse_transform(pcaed2)
    orig.make_stacked()
    assert np.allclose(feats.stacked_features, orig.stacked_features)

    assert BagPCA(k=5, randomize=True).fit_transform(bags).dim == 5

    assert_raises(TypeError, lambda: BagPCA(randomize=True))
    assert_raises(TypeError, lambda: BagPCA(mle_components=True, k=12))
    assert BagPCA(mle_components=True)



################################################################################

if __name__ == '__main__':
    import nose
    nose.main()



from __future__ import division, print_function

from copy import deepcopy
import warnings

import numpy as np
from sklearn.externals.six import iteritems, string_types
from sklearn.externals.six.moves import xrange

from .utils import as_integer_type


class Features(object):
    '''
    A wrapper class for storing bags of features. (A *bag* is a set of feature
    vectors corresponding to a single "object.")

    Supports storing data in two major ways:

    - As a list of pointers to a numpy array per bag. This is the default,
      because it usually doesn't require copying all of your data. Note that
      the sub-arrays are not enforced to be row-major or even contiguous.

    - As a single big row-major array. This lets you do certain things more
      easily (e.g. run PCA).

    The main usage API is the same for both versions; you can distinguish them
    with the `stacked` property, and convert from pointers to stacked with the
    `make_stacked()` method.

    Supports the following operations:

    * ``len(features)`` gives the number of bags.
    * ``for bag in features:`` loops over bags.
    * ``features[4]`` gives the fifth bag.
    * ``features[[4, 8, 7]]`` makes a new Features object with only the \
       passed indices, preserving metadata.
    * ``feats1 == feats2`` checks that all of the features `and metadata` are \
      the same.
    * ``feats1 + feats2`` concatenates the two Features objects. Metadata is \
      preserved if both features have that key, thrown out if not.


    Parameters
    ----------
    bags : list of numpy arrays, single array, or Features object
        The feature data. If a list of numpy arrays, should be one array per 
        bag, each of shape [n_pts, dim], where dim is consistent between bags
        but n_pts need not be (though it cannot ever be 0). If a single numpy
        array, it should be of shape [sum(n_pts), dim] and contain the features
        from the first bag, then the next bag, .... In this case you must also
        pass n_pts. If a Features object, "copies" it (but only actually copies
        any data if ``copy=True``).

    n_pts : array-like of positive integers, only if bags is a single array
        If bags is passed as a single array, a list of positive integers
        defining the size of each bag.

    stack : boolean, optional, default False
        If true, stack the features. Otherwise, only stack them if a stacked
        array of features is passed in.

    copy : boolean, optional, default False
        If true, always make a copy of the data (so that direct modifications)
        don't modify the original arrays. If false, make a copy only if
        necessary (i.e. stack=True for an unstacked argument).

    bare : boolean, optional, default False
        If true, and ``bags`` is a Features instance, don't include its
        metadata.

    any other keyword argument : array-like with first dimension num_bags
        Metadata for each bag. Just stored along with the features, nothing
        in particular done with it. (If ``bags`` is a Features instance, its
        metadata is merged with any keyword arguments, with keywords taking
        precedence.)


    Attributes
    ----------
    features : list of arrays of shape ``[n_pts[i], dim]``
        A list of the contained features.
        If ``stacked``, each array is a slice of ``stacked_features``.

    stacked : boolean
        Whether the features are stacked.

    stacked_features : array of shape ``[sum(n_pts), dim]``
        All of the features, concatenated together. Only present if ``stacked``.

    n_pts : integer array of shape ``[len(self)]``
        The number of points in each bag.

    meta : dictionary mapping strings to arrays of shape ``[len(self)]``
        The stored metadata. ``meta['foo']`` is also accessible as ``self.foo``.
    '''

    def __init__(self, bags, n_pts=None, stack=False, copy=False, bare=False,
                 **meta):
        if isinstance(bags, Features):
            if n_pts is not None:
                raise TypeError("can't pass n_pts if copying a Features object")

            oth = bags
            if oth.stacked:
                bags = oth.stacked_features
                n_pts = oth.n_pts
            else:
                bags = oth.features
                n_pts = None
            if not bare:
                for k, v in iteritems(oth.meta):
                    meta.setdefault(k, v)


        if isinstance(bags, np.ndarray) and bags.ndim == 2:
            if n_pts is None:
                raise TypeError("must pass n_pts if passing stacked array of "
                                "features")

            n_pts = np.asarray(n_pts)
            if n_pts.ndim != 1:
                raise TypeError("n_pts must be 1-dimensional")
            if n_pts.size == 0:
                raise TypeError("must have at least one bag")
            if np.any(n_pts <= 0):
                raise TypeError("n_pts must all be positive")
            try:
                n_pts = as_integer_type(n_pts)
            except ValueError:
                raise TypeError("n_pts must be an array of integers.")

            bags = np.array(bags, order='C', copy=copy)
            if bags.ndim != 2 or bags.shape[0] != np.sum(n_pts):
                raise TypeError("bags must have shape sum(n_pts) x dim")
            if bags.shape[1] == 0:
                raise TypeError("bags must have dimension > 0")
            dim = bags.shape[1]

            self.stacked = True
            self.n_pts = n_pts
            self.stacked_features = bags
            self._boundaries = bounds = np.r_[0, np.cumsum(n_pts)]

            self.features = np.empty(len(n_pts), object)
            self.features[:] = [bags[bounds[i-1]:bounds[i]]
                                for i in xrange(1, len(bounds))]

        else:
            if n_pts is not None:
                raise TypeError("n_pts should only be passed if bags is a "
                                "single stacked array")

            dim = None
            dtype = None
            new_bags = np.empty(len(bags), dtype=object)
            n_pts = np.empty(len(bags), dtype=int)
            for i, bag in enumerate(bags):
                a = np.array(bag, copy=copy)

                if a.ndim == 1:
                    a = a[None, :]
                if a.ndim != 2:
                    raise TypeError("bag {} not two-dimensional".format(i))

                if dim is None:
                    dim = a.shape[1]
                elif a.shape[1] != dim:
                    msg = "bags' second dimension must be consistent: " \
                          "{} is {}, expected {}"
                    raise TypeError(msg.format(i, a.shape[1], dim))

                if dtype is None:
                    dtype = a.dtype
                    if dtype.kind not in 'fiu':
                        msg = "can't handle features of type {}"
                        raise TypeError(msg.format(a.dtype.name))
                elif a.dtype != dtype:
                    msg = "bags' dtype is inconsistent: {} is {}, expected {}"
                    raise TypeError(msg.format(i, a.dtype.name, dtype.name))

                if a.shape[0] == 0:
                    raise TypeError("bag {} has no points".format(i))

                new_bags[i] = a
                n_pts[i] = a.shape[0]

            self.stacked = False
            self.n_pts = n_pts
            self.features = new_bags
            try:
                del self._boundaries
            except AttributeError:
                pass
            try:
                del self.stacked_features
            except AttributeError:
                pass

            if stack:
                self.make_stacked()

        # handle metadata
        self.meta = {}
        for name, val in iteritems(meta):
            if len(val) != len(n_pts):
                msg = "Have {} bags but {} values for {}"
                raise ValueError(msg.format(len(n_pts), len(val), name))

            val = np.array(val, copy=copy)
            self.meta[name] = val
            if hasattr(self, name):
                msg = "Features already has an attribute named '{}'; won't " \
                      "be accessible as an attribute"
                warnings.warn(msg.format(name))
            else:
                setattr(self, name, val)

    def make_stacked(self):
        "If unstacked, convert to stacked. If stacked, do nothing."
        if self.stacked:
            return

        self._boundaries = bounds = np.r_[0, np.cumsum(self.n_pts)]
        self.stacked_features = stacked = np.vstack(self.features)
        self.features = np.array(
            [stacked[bounds[i-1]:bounds[i]] for i in xrange(1, len(bounds))],
            dtype=object)
        self.stacked = True

    ############################################################################
    ## Properties to get at basic metadata

    @property
    def total_points(self):
        "The total number of points in all bags."
        return self.n_pts.sum()

    @property
    def dim(self):
        "The dimensionality of the features."
        return self.features[0].shape[1]

    @property
    def dtype(self):
        "The data type of the feature vectors."
        return self.features[0].dtype

    ############################################################################
    ## Copying / pickling utilities

    def copy(self, stack=False, copy_meta=False, memo=None):
        '''
        Copies the Feature object. Makes a copy of the features array.

        Parameters
        ----------
        stack : boolean, optional, default False
            Whether to stack the copy if this one is unstacked.

        copy_meta : boolean, optional, default False
            Also copy the metadata. If False, metadata in both points to the
            same object.
        '''
        if self.stacked:
            fs = deepcopy(self.stacked_features, memo)
            n_pts = self.n_pts.copy()
        elif stack:
            fs = np.vstack(self.features)
            n_pts = self.n_pts.copy()
        else:
            fs = deepcopy(self.features, memo)
            n_pts = None

        meta = deepcopy(self.meta, memo) if copy_meta else self.meta
        return Features(fs, n_pts, copy=False, **meta)

    def __copy__(self):
        return self.copy(stack=False, copy_meta=False)

    def __deepcopy__(self, memo=None):
        return self.copy(stack=False, copy_meta=True, memo=memo)

    def __getstate__(self):
        if self.stacked:
            return (self.stacked_features, self.n_pts, self.meta)
        else:
            return (self.features, None, self.meta)

    def __setstate__(self, state):
        feats, n_pts, meta = state
        self.__init__(feats, n_pts, **meta)

    ############################################################################ 
    ## General magic methods for basic behavior

    __hash__ = None

    def __eq__(self, oth):
        if self is oth:
            return True
        elif isinstance(oth, Features):
            return (len(self) == len(oth) and
                    set(self.meta) == set(oth.meta) and
                    all(np.all(self_b == oth_b)
                        for self_b, oth_b in zip(self, oth)) and
                    all(np.all(self.meta[k] == oth.meta[k])
                        for k in self.meta))
        elif self.meta:
            return False
        else:
            return (len(self) == len(oth) and 
                    all(np.all(self_b == oth_b)
                        for self_b, oth_b in zip(self, oth)))

    def __ne__(self, oth):
        return not (self == oth)

    def __repr__(self):
        s = '<Features: {:,} bags with {} {}-dimensional points ({:,} total)>'
        min_p = self.n_pts.min()
        max_p = self.n_pts.max()
        if min_p == max_p:
            pts = "{:,}".format(min_p)
        else:
            pts = '{:,} to {:,}'.format(min_p, max_p)
        return s.format(len(self), pts, self.dim, self.total_points)

    def __len__(self):
        return self.n_pts.size

    def __iter__(self):
        return iter(self.features)

    def __getitem__(self, key):
        if (isinstance(key, string_types) or
                (isinstance(key, (tuple, list)) and
                 any(isinstance(x, string_types) for x in key))):
            msg = "Features indexing only subsets rows, but got {!r}"
            raise TypeError(msg.format(key))

        if np.isscalar(key):
            return self.features[key]
        else:
            return type(self)(self.features[key], copy=False, stack=False,
                              **{k: v[key] for k, v in iteritems(self.meta)})

    def __add__(self, oth):
        if isinstance(oth, Features):
            meta = {k: np.r_[self.meta[k], oth.meta[k]]
                    for k in self.meta if k in oth.meta}
            oth_features = oth.features
        elif isinstance(oth, list):
            meta = {}
            oth_features = np.empty(len(oth), object)
            oth_features[:] = oth
        else:
            return NotImplemented

        return Features(np.r_[self.features, oth_features],
                        stack=False, copy=True, **meta)

    def __radd__(self, oth):
        if isinstance(oth, list):
            oth_features = np.empty(len(oth), object)
            oth_features[:] = oth
        else:
            return NotImplemented
        return Features(np.r_[oth_features, self.features],
                        stack=False, copy=True)

    ############################################################################
    ## Others

    def bare(self):
        "Make a Features object with no metadata; points to the same features."
        if not self.meta:
            return self
        elif self.stacked:
            return Features(self.stacked_features, self.n_pts, copy=False)
        else:
            return Features(self.features, copy=False)


def as_features(X, stack=False, bare=False):
    '''
    Returns a version of X as a :class:`Features` object.

    Parameters
    ----------
    stack : boolean, default False
        Make a stacked version of X. Note that if X is a features object,
        this will stack it in-place, since that's usually what you want.
        (If not, just use the :class:`Features` constructor instead.)

    bare : boolean, default False
        Return a bare version of X (no metadata).

    Returns
    -------
    feats : :class:`Features`
        A version of X. If X is already a :class:`Features` object, the original
        X may be returned, depending on the arguments.
    '''
    if isinstance(X, Features):
        if stack:
            X.make_stacked()
        return X.bare() if bare else X
    return Features(X, stack=stack, bare=bare)



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin

from ..features import as_features

class BagMean(BaseEstimator, TransformerMixin):
    '''
    Transforms a list of features or :class:`Features` object by just taking
    the mean of each entry.

    Examples
    --------
    If you have a kernel where you can compute an (approximate?) explicit
    embedding, you can compute an explicit embedding of
    :class:`skl_groups.kernels.MeanMapKernel` by first doing the kernel
    embedding and then taking the mean.
    Using :class:`sklearn.kernel_approximation.RBFSampler`::

        rbf_mmk = Pipeline([
            ('rbf', BagPreprocesser(RBFSampler(gamma=g, n_components=100))),
            ('mean', BagMean()),
        ])
    '''

    def __init__(self):
        pass

    def fit(self, X, y=None):
        "Do nothing; this is a stateless transformer."
        return self

    def transform(self, X):
        '''
        Transform a list of bag features into a matrix of its mean features.

        Parameters
        ----------
        X : :class:`skl_groups.features.Features` or list of bag feature arrays
            Data to transform.

        Returns
        -------
        X_new : array, shape ``[len(X), X.dim]``
            X transformed into its means.
        '''
        X = as_features(X)
        return np.vstack([np.mean(bag, axis=0) for bag in X])



from __future__ import division

import os
import sys

import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.cluster import KMeans, MiniBatchKMeans
from sklearn.externals.six.moves import xrange
from nose.tools import assert_raises

if __name__ == '__main__':
    # make this copy of skl_groups importable
    _this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(_this_dir)))

from skl_groups.features import Features
from skl_groups.preprocessing import BagMinMaxScaler
from skl_groups.summaries import BagOfWords, BagMean, L2DensityTransformer


################################################################################

def test_mean():
    dim = 5
    n_bags = 50
    np.random.seed(42)
    bags = [np.random.randn(np.random.randint(30, 100), dim)
            for _ in xrange(n_bags)]

    meaned = BagMean().fit_transform(bags)
    assert meaned.shape == (n_bags, dim)
    assert np.allclose(meaned[3], np.mean(bags[3], axis=0))


def test_bagofwords_basic():
    n_codewords = 10
    dim = 5
    kmeans = KMeans(n_clusters=n_codewords, max_iter=100, n_init=3,
                    random_state=47)
    bow = BagOfWords(kmeans)

    np.random.seed(42)
    bags = [np.random.randn(np.random.randint(30, 100), dim)
            for _ in xrange(50)]

    bowed = bow.fit_transform(bags)
    assert bowed.shape == (len(bags), n_codewords)
    assert bow.codewords_.shape == (n_codewords, dim)
    assert np.all(bowed >= 0)
    assert np.all(np.sum(bowed, 1) == [b.shape[0] for b in bags])

    bow.fit(Features(bags))
    bowed2 = bow.transform(bags)
    assert np.all(bowed == bowed2)
    assert bow.codewords_.shape == (n_codewords, dim)

    minikmeans = MiniBatchKMeans(n_clusters=n_codewords, max_iter=100,
                                 random_state=47)
    minibow = BagOfWords(minikmeans)
    assert_raises(AttributeError, lambda: minibow.transform(bags))
    minibowed = minibow.fit_transform(bags)
    assert minibowed.shape == bowed.shape
    assert np.all(bowed >= 0)
    assert np.all(np.sum(bowed, 1) == [b.shape[0] for b in bags])


def test_l2density_basic():
    dim = 3
    bags = [np.random.randn(np.random.randint(30, 100), dim)
            for _ in xrange(50)]
    pipe = Pipeline([
        ('scale', BagMinMaxScaler([0, 1])),
        ('density', L2DensityTransformer(15)),
    ])
    l2ed = pipe.fit_transform(bags)

    assert np.all(np.isfinite(l2ed))
    # ||x - y||^2 = <x, x> - 2 <x, y> + <y, y>
    K = l2ed.dot(l2ed.T)
    row_norms_sq = np.diagonal(K)
    l2_dist_sq = row_norms_sq[:, None] - 2 * K + row_norms_sq[None, :]
    assert np.min(row_norms_sq) > 0
    assert np.min(l2_dist_sq) >= 0

    assert_raises(ValueError, lambda: L2DensityTransformer(10, basis='foo'))

    t = L2DensityTransformer(10)
    assert_raises(AttributeError, lambda: t.transform(bags))
    t.fit(dim)
    t.transform(BagMinMaxScaler([0, 1]).fit_transform(bags))
    assert_raises(ValueError, lambda: t.transform([b[:, :2] for b in bags]))
    assert_raises(ValueError, lambda: t.transform(bags))
    t.basis = 'haha snuck my way in'
    assert_raises(ValueError, lambda: t.transform(bags))


################################################################################

if __name__ == '__main__':
    import nose
    nose.main()



from __future__ import division

import numpy as np

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.metrics.pairwise import pairwise_kernels
from sklearn.externals.six.moves import xrange as range

from ..features import as_features


class MeanMapKernel(BaseEstimator, TransformerMixin):
    '''
    Computes a kernel between bags as the mean pairwise evaluation of a kernel
    between the elements of the bags.

    Also known as the maximum mean discrepancy (MMD) when used for a hypothesis
    test for whether two samples come from the same distribution.

    Parameters
    ----------
    kernel : string or callable, default="rbf"
        Kernel map to be approximated. A callable should accept two arguments
        and the keyword arguments passed to this object as kernel_params, and
        should return a floating point number.

    gamma : float, default=None
        Gamma parameter for the RBF, polynomial, exponential chi2 and
        sigmoid kernels. Interpretation of the default value is left to
        the kernel; see :mod:`sklearn.metrics.pairwise`.
        Ignored by other kernels.

    degree : float, default=3
        Degree of the polynomial kernel. Ignored by other kernels.

    coef0 : float, default=1
        Zero coefficient for polynomial and sigmoid kernels.
        Ignored by other kernels.

    kernel_params : mapping of string to any, optional
        Additional parameters (keyword arguments) for kernel function passed
        as callable object.

    See also
    --------
    :class:`skl_groups.summaries.BagMean` is the equivalent if you have an
    explicit kernel map.
    '''

    _pairwise = False
    _pairwise_output = True

    def __init__(self, kernel="rbf", gamma=None, coef0=1, degree=3,
                 kernel_params=None, n_components=100):
        self.kernel = kernel
        self.gamma = gamma
        self.coef0 = coef0
        self.degree = degree
        self.kernel_params = kernel_params

    def _get_kernel_params(self):
        params = self.kernel_params
        if params is None:
            params = {}
        if not callable(self.kernel):
            params['gamma'] = self.gamma
            params['degree'] = self.degree
            params['coef0'] = self.coef0

        return params

    def fit(self, X, y=None):
        '''
        Specify the data to which kernel values should be computed.

        Parameters
        ----------
        X : list of arrays or :class:`skl_groups.features.Features`
            The bags to compute "to".
        '''
        self.features_ = as_features(X, stack=True, bare=True)
        # TODO: could precompute things like squared norms if kernel == "rbf".
        # Probably should add support to sklearn instead of hacking it here.
        return self

    def transform(self, X):
        '''
        Compute kernels from X to :attr:`features_`.

        Parameters
        ----------
        X : list of arrays or :class:`skl_groups.features.Features`
            The bags to compute "from". Must have same dimension as
            :attr:`features_`.

        Returns
        -------
        K : array of shape ``[len(X), len(features_)]``
            The kernel evaluations from X to :attr:`features_`.
        '''

        X = as_features(X, stack=True, bare=True)
        Y = self.features_

        if X.dim != Y.dim:
            raise ValueError("MMK transform got dimension {} but had {} at fit"
                             .format(X.dim, Y.dim))

        pointwise = pairwise_kernels(X.stacked_features, Y.stacked_features,
                                     metric=self.kernel,
                                     filter_params=True,
                                     **self._get_kernel_params())

        # TODO: is there a way to do this without a Python loop?
        K = np.empty((len(X), len(Y)))
        for i in range(len(X)):
            for j in range(len(Y)):
                K[i, j] = pointwise[X._boundaries[i]:X._boundaries[i+1],
                                    Y._boundaries[j]:Y._boundaries[j+1]].mean()

        return K



import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA, RandomizedPCA
from sklearn.preprocessing import StandardScaler, Normalizer
from sklearn.utils import check_array

from .features import Features, as_features

# TODO: support inplace transformations here

class BagPreprocesser(BaseEstimator, TransformerMixin):
    '''
    Applies a preprocessing estimator to each column of the bags, independently.
    This is a reasonable thing to do for some cases
    (especially :class:`sklearn.preprocessing.StandardScaler`,
    :class:`sklearn.preprocessing.MinMaxScaler`,
    :class:`sklearn.decomposition.PCA`).
    It's not a reasonable thing to do for things that rely on interactions
    between points, or that change the number of output points. (Changing
    the dimension is okay.)

    Parameters
    ----------
    transformer : an sklearn transformer
        The transformer to apply to the stacked features. Must return the
        same number of features.
    '''
    def __init__(self, transformer):
        t = transformer
        if (not (hasattr(t, "fit") or hasattr(t, 'fit_transform')) or
                not hasattr(t, "transform")):
            raise TypeError("The transformer doesn't have appropriate methods.")
        self.transformer = t

    def _gather_outputs(self, old, new):
        if new.shape[0] != old.total_points:
            msg = "Transformer changed number of points from {} to {}"
            raise ValueError(msg.format(old.total_points, new.shape[0]))
        return Features(new, old.n_pts, **old.meta)

    def fit(self, X, y=None, **params):
        '''
        Fit the transformer on the stacked points.

        Parameters
        ----------
        X : :class:`Features` or list of arrays of shape ``[n_samples[i], n_features]``
            Training set. If a Features object, it will be stacked.

        any other keyword argument :
            Passed on as keyword arguments to the transformer's ``fit()``.
        '''
        X = as_features(X, stack=True)
        self.transformer.fit(X.stacked_features, y, **params)
        return self

    def transform(self, X, **params):
        '''
        Transform the stacked points.

        Parameters
        ----------
        X : :class:`Features` or list of bag feature arrays
            New data to transform.

        any other keyword argument :
            Passed on as keyword arguments to the transformer's ``transform()``.

        Returns
        -------
        X_new : :class:`Features`
            Transformed features.
        '''
        X = as_features(X, stack=True)
        X_new = self.transformer.transform(X.stacked_features, **params)
        return self._gather_outputs(X, X_new)

    def fit_transform(self, X, y=None, **params):
        '''
        Fit and transform the stacked points.

        Parameters
        ----------
        X : :class:`Features` or list of bag feature arrays
            Data to train on and transform.

        any other keyword argument :
            Passed on as keyword arguments to the transformer's ``transform()``.

        Returns
        -------
        X_new : :class:`Features`
            Transformed features.
        '''
        X = as_features(X, stack=True)
        X_new = self.transformer.fit_transform(X.stacked_features, y, **params)
        return self._gather_outputs(X, X_new)

    def inverse_transform(self, X, **params):
        '''
        Transform data back to its original space, i.e., return an input
        X_original whose transform would (maybe approximately) be X.

        Parameters
        ----------
        X : :class:`Features` or list of bag feature arrays
            Data to train on and transform.

        any other keyword argument :
            Passed on as keyword arguments to the transformer's 
            ``inverse_transform()``.

        Returns
        -------
        X_original : :class:`Features`
        '''
        X = as_features(X, stack=True)
        Xo = self.transformer.inverse_transform(X.stacked_features, **params)
        return self._gather_outputs(X, Xo)


class BagStandardizer(BagPreprocesser):
    '''
    Standardizes each feature dimension to have zero mean and unit variance,
    regardless of the bag it falls into.

    This is just :class:`BagPreprocesser` with
    :class:`sklearn.preprocessing.StandardScaler`.
    '''
    def __init__(self):
        super(BagStandardizer, self).__init__(StandardScaler())


class MinMaxScaler(BaseEstimator, TransformerMixin):
    """Standardizes features by scaling each feature to a given range.

    This estimator scales and translates each feature individually such
    that it is in the given range on the training set, i.e. between
    zero and one.

    The standardization is given by::
        X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
        X_scaled = X_std * (max - min) + min

    where min, max = feature_range.

    This standardization is often used as an alternative to zero mean,
    unit variance scaling.

    Notes
    -----

    This is a version of :class:`sklearn.preprocessing.MinMaxScaler`
    with support for truncation added. It's been
    `proposed <https://github.com/scikit-learn/scikit-learn/pull/3342>`_
    for inclusion in scikit-learn, but is not yet in there.

    Parameters
    ----------
    feature_range: tuple (min, max), default=(0, 1)
        Desired range of transformed data.

    copy : boolean, optional, default is True
        Set to False to perform inplace row normalization and avoid a
        copy (if the input is already a numpy array).

    truncate : boolean, optional, default is False
        If True, :meth:`transform` will truncate any inputs that lie outside
        the min/max of the values passed to :meth:`fit` to lie on the ends
        of feature_range. Normally, the transform of these points will be
        outside feature_range.

    fit_feature_range : None or tuple (min, max), default None
        If not None, :meth:`fit` will actually rescale such that the passed
        features all lie within fit_feature_range rather than just
        feature_range. This is useful when truncate is True, to give
        some "wiggle room" before it starts truncating. Otherwise it just
        effectively overrides feature_range.

    Attributes
    ----------
    `min_` : ndarray, shape (n_features,)
        Per feature adjustment for minimum.

    `scale_` : ndarray, shape (n_features,)
        Per feature relative scaling of the data.
    """

    def __init__(self, feature_range=(0, 1), copy=True, truncate=False,
                 fit_feature_range=None):
        self.feature_range = feature_range
        self.copy = copy
        self.truncate = truncate
        self.fit_feature_range = fit_feature_range

    def fit(self, X, y=None):
        """Compute the minimum and maximum to be used for later scaling.

        Parameters
        ----------
        X : array-like, shape [n_samples, n_features]
            The data used to compute the per-feature minimum and maximum
            used for later scaling along the features axis.
        """
        X = check_array(X, copy=self.copy,
                        dtype=[np.float64, np.float32, np.float16, np.float128])

        feature_range = self.feature_range
        if feature_range[0] >= feature_range[1]:
            raise ValueError("Minimum of desired feature range must be smaller"
                             " than maximum. Got %s." % str(feature_range))
        if self.fit_feature_range is not None:
            fit_feature_range = self.fit_feature_range
            if fit_feature_range[0] >= fit_feature_range[1]:
                raise ValueError("Minimum of desired (fit) feature range must "
                                 "be smaller than maximum. Got %s."
                                 % str(feature_range))
            if (fit_feature_range[0] < feature_range[0] or
                    fit_feature_range[1] > feature_range[1]):
                raise ValueError("fit_feature_range must be a subset of "
                                 "feature_range. Got %s, fit %s."
                                 % (str(feature_range),
                                    str(fit_feature_range)))
            feature_range = fit_feature_range

        data_min = np.min(X, axis=0)
        data_range = np.max(X, axis=0) - data_min
        # Do not scale constant features
        data_range[data_range == 0.0] = 1.0
        self.scale_ = (feature_range[1] - feature_range[0]) / data_range
        self.min_ = feature_range[0] - data_min * self.scale_
        self.data_range = data_range
        self.data_min = data_min
        return self

    def transform(self, X):
        """Scaling features of X according to feature_range.

        Parameters
        ----------
        X : array-like with shape [n_samples, n_features]
            Input data that will be transformed.
        """
        X = check_array(X, copy=self.copy)
        X *= self.scale_
        X += self.min_
        if self.truncate:
            np.maximum(self.feature_range[0], X, out=X)
            np.minimum(self.feature_range[1], X, out=X)
        return X

    def inverse_transform(self, X):
        """Undo the scaling of X according to feature_range.

        Note that if truncate is true, any truncated points will not
        be restored exactly.

        Parameters
        ----------
        X : array-like with shape [n_samples, n_features]
            Input data that will be transformed.
        """
        X = check_array(X, copy=self.copy)
        X -= self.min_
        X /= self.scale_
        return X



class BagMinMaxScaler(BagPreprocesser):
    '''
    Linearly scales each feature dimension to lie within the given range, for
    example [0, 1].

    This is just :class:`BagPreprocesser` with :class:`MinMaxScaler`.

    Parameters
    ----------
    feature_range : tuple (min, max), default = (0, 1)
        Desired range of the transformed data.
    '''
    def __init__(self, feature_range=(0, 1), truncate=False,
                 fit_feature_range=None):
        super(BagMinMaxScaler, self).__init__(MinMaxScaler(
            feature_range=feature_range, truncate=truncate,
            fit_feature_range=fit_feature_range))


class BagNormalizer(BagPreprocesser):
    '''
    Normalizes each sample individually to have unit norm (l1 or l2).

    This is just :class:`BagPreprocesser` with
    :class:`sklearn.preprocessing.Normalizer`.

    Parameters
    ----------
    norm : 'l1' or 'l2', optional, default 'l2'
        The norm to use to normalize each nonzero sample.
    '''
    def __init__(self, norm='l2'):
        super(BagNormalizer, self).__init__(Normalizer(norm))


DEFAULT_VARFRAC = 0.7
class BagPCA(BagPreprocesser):
    '''
    Runs principal components analysis to reduce the dimensionality of the
    features.

    This is just :class:`BagPreprocesser` with
    either :class:`sklearn.decomposition.PCA`
    or :class:`sklearn.decomposition.RandomizedPCA`.

    Parameters
    ----------
    k : int, optional
        The dimensionality to reduce to.

    mle_components : boolean, optional, default False
        Use Minka's MLE for determining the number of components.

    varfrac : float in (0, 1], optional, default 0.7
        Use enough components to cover this fraction of the variance.
        Only one of {k, mle_components, varfrac} can be passed.

    randomize : boolean, optional, default False
        Use a randomized PCA. This can be faster and less memory-intensive for
        large inputs, but is approximate and requires specifying an explicit
        number of components.

    whiten : boolean, optional, default False
        Whether to whiten the outputs, by dividing the components by the
        singular values. This removes some information, but makes the variance
        of the outputs the identity matrix.
    '''
    def __init__(self, k=None, mle_components=False, varfrac=None,
                 randomize=False, whiten=False):
        n_specs = sum(1 for x in [k, mle_components, varfrac] if x)
        if n_specs > 1:
            msg = "can't specify number of components in more than one way"
            raise TypeError(msg)
        if n_specs == 0:
            varfrac = DEFAULT_VARFRAC

        if randomize:
            if k is None:
                raise TypeError("can't do random PCA without a specific k")
            pca = RandomizedPCA(k, whiten=whiten)
        else:
            if k is not None:
                n_components = k
            elif mle_components:
                n_components = 'mle'
            elif varfrac is not None:
                n_components = varfrac
            pca = PCA(n_components, whiten=whiten)
        super(BagPCA, self).__init__(pca)



'''
Methods that assume feature sets are samples from some probability distribution
and estimate divergences, distances, and so on between those distributions.
'''
from ..utils import NoImportStub

try:
    from .knn import KNNDivergenceEstimator
except ImportError as e:
    import sys
    KNNDivergenceEstimator = NoImportStub(sys.exc_info())



from copy import copy

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.externals.six.moves import xrange
from sklearn.utils.validation import check_is_fitted

from ..features import as_features

# TODO: support vlfeat's kmeans algorithms
# (to separate fit() and predict(), need to refactor vlfeat-ctypes a bit)

class BagOfWords(BaseEstimator, TransformerMixin):
    '''
    Transforms feature sets into the bag of words representation:

    1. Run k-means on the set of all points from all bags, to get codewords.
    2. Represent each bag by the count of points assigned to each codeword.

    This approach is common in computer vision and, when the number of codewords
    is tuned correctly, can provide a good single summary vector of the
    distribution of inputs for many types of learning tasks.

    It is often helpful to either take the square root of the output vectors
    or compare them with :func:`sklearn.metrics.pairwise.chi2_kernel`, which
    have similar effects and both approximate a more realistic non-iid model
    [1]_.


    Parameters
    ----------
    kmeans : KMeans estimator
        Object to run k-means with; :class:`sklearn.cluster.KMeans` or
        :class:`sklearn.cluster.MiniBatchKMeans` are most likely.
        Compared to a standard clustering setup, you want a much higher
        ``n_clusters`` (depending on the size and dimensionality, between maybe
        tens and thousands), lowish ``max_iter`` (maybe 100; we don't care so
        much about the exact solution, just that it has a reasonable spread),
        and lower ``n_init`` (same reason). The object should not yet be fit;
        we'll copy it, not modify it.

    Attributes
    ----------
    `kmeans_fit_` : KMeans estimator
        A fit version of the `kmeans` parameter.

    References
    ----------
    .. [1] R. Cinbis, J. Verbeek, & C. Schmid (2012).
           Image categorization using Fisher kernels on non-iid image models.
           In Computer Vision and Pattern Recognition (CVPR).
    '''
    def __init__(self, kmeans):
        self.kmeans = kmeans

    def _check_fitted(self):
        if not hasattr(self, "kmeans_fit_"):
            raise AttributeError("Model has not been trained yet.")
        check_is_fitted(self.kmeans_fit_, 'cluster_centers_')

    @property
    def n_codewords(self):
        "The number of codewords used."
        return self.kmeans.n_clusters

    @property
    def codewords_(self):
        "The selected codewords; shape [n_codewords, n_features]."
        self._check_fitted()
        return self.kmeans_fit_.cluster_centers_

    def _group_assignments(self, X, assignments):
        X_new = np.empty((len(X), self.n_codewords), dtype=np.int32)
        for i in xrange(len(X)):
            start = X._boundaries[i]
            end = X._boundaries[i + 1]

            X_new[i, :] = np.bincount(assignments[start:end],
                                      minlength=self.n_codewords)
        return X_new

    def fit(self, X, y=None):
        '''
        Choose the codewords based on a training set.

        Parameters
        ----------
        X : :class:`skl_groups.features.Features` or list of arrays of shape ``[n_samples[i], n_features]``
            Training set. If a Features object, it will be stacked.
        '''
        self.kmeans_fit_ = copy(self.kmeans)
        X = as_features(X, stack=True)
        self.kmeans_fit_.fit(X.stacked_features) 
        return self

    def transform(self, X):
        '''
        Transform a list of bag features into its bag-of-words representation.

        Parameters
        ----------
        X : :class:`skl_groups.features.Features` or list of bag feature arrays
            New data to transform.

        Returns
        -------
        X_new : integer array, shape [len(X), kmeans.n_clusters]
            X transformed into the new space.
        '''
        self._check_fitted()
        X = as_features(X, stack=True)
        assignments = self.kmeans_fit_.predict(X.stacked_features)
        return self._group_assignments(X, assignments)

    def fit_transform(self, X):
        '''
        Compute clustering and transform a list of bag features into its
        bag-of-words representation. Like calling fit(X) and then transform(X),
        but more efficient.

        Parameters
        ----------
        X : :class:`skl_groups.features.Features` or list of bag feature arrays
            New data to transform.

        Returns
        -------
        X_new : integer array, shape [len(X), kmeans.n_clusters]
            X transformed into the new space.
        '''
        X = as_features(X, stack=True)
        self.kmeans_fit_ = copy(self.kmeans)
        assignments = self.kmeans_fit_.fit_predict(X.stacked_features) 
        return self._group_assignments(X, assignments)



from __future__ import division

import os
import sys

import numpy as np
from sklearn.externals.six.moves import range
from sklearn.metrics.pairwise import pairwise_kernels
from sklearn.utils.testing import assert_almost_equal
from nose.tools import assert_raises

if __name__ == '__main__':
    # make this copy of skl_groups importable
    _this_dir = os.path.dirname(os.path.abspath(__file__))
    sys.path.insert(0, os.path.dirname(os.path.dirname(_this_dir)))

from skl_groups.kernels import MeanMapKernel


################################################################################

def test_mmk():
    bags = [np.random.normal(size=(np.random.randint(10, 100), 10))
            for _ in range(20)]

    res = MeanMapKernel(gamma=2.38).fit_transform(bags)
    for i in range(20):
        for j in range(20):
            exp = pairwise_kernels(bags[j], bags[i], metric='rbf', gamma=2.38)
            assert_almost_equal(res[i, j], exp.mean(),
                                err_msg="({} to {})".format(i, j))

    res = MeanMapKernel(kernel='linear').fit(bags[:5]).transform(bags[-2:])
    for i in range(5):
        for j in range(18, 20):
            exp = pairwise_kernels(bags[j], bags[i], metric='linear')
            assert_almost_equal(res[j - 18, i], exp.mean(),
                                err_msg="({} to {})".format(i, j))

    # fails on wrong dimension
    assert_raises(
        ValueError,
        lambda:MeanMapKernel().fit(bags).transform([np.random.randn(20, 8)]))


################################################################################

if __name__ == '__main__':
    import nose
    nose.main()



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import versiontools_support

DESCRIPTION = "Addon to scikit-learn for handling set-based data."
with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()
NAME = "skl-groups"
AUTHOR = MAINTAINER = "Dougal J. Sutherland"
AUTHOR_EMAIL = MAINTAINER_EMAIL = "dsutherl@cs.cmu.edu"
URL = DOWNLOAD_URL = "https://github.com/dougalsutherland/skl-groups"
LICENSE = "BSD"
CLASSIFIERS = [
    "Development Status :: 2 - Pre-Alpha",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.3",
    "Programming Language :: Python :: 3.4",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Topic :: Scientific/Engineering :: Image Recognition",
    "Topic :: Scientific/Engineering :: Information Analysis",
]


setup(
    name=NAME,
    version=':versiontools:skl_groups',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    url=URL,
    download_url=DOWNLOAD_URL,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    packages=[
        'skl_groups',
        'skl_groups.divergences',
        'skl_groups.kernels',
        'skl_groups.summaries',
        'skl_groups.tests',
    ],
    install_requires=['scikit-learn >= 0.17'],
    extras_require={
        'knn_divs': ['cyflann >= 0.1.15'],
    },
    zip_safe=False,
)



# Copyright (C) 2012 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of versiontools.
#
# versiontools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# versiontools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with versiontools.  If not, see <http://www.gnu.org/licenses/>.

"""
versiontools.versiontools_support
=================================

A small standalone module that allows any package to use versiontools.

Typically you should copy this file verbatim into your source distribution.

Historically versiontools was depending on a exotic feature of setuptools to
work. Setuptools has so-called setup-time dependencies, that is modules that
need to be downloaded and imported/interrogated for setup.py to run
successfully. Versiontools supports this by installing a handler for the
'version' keyword of the setup() function.

This approach was always a little annoying as this setuptools feature is rather
odd and very few other packages made any use of it. In the future the standard
tools for python packaging (especially in python3 world) this feature may be
removed or have equivalent thus rendering versiontools completely broken.

Currently the biggest practical issue is the apparent inability to prevent
setuptools from downloading packages designated as setup_requires. This is
discussed in this pip issue: https://github.com/pypa/pip/issues/410

To counter this issue I've redesigned versiontools to be a little smarter. The
old mode stays as-is for compatibility. The new mode works differently, without
the need for using setup_requires in your setup() call. Instead it requires
each package that uses versiontools to ship a verbatim copy of this module and
to import it in their setup.py script. This module helps setuptools find
package version in the standard PKG-INFO file that is created for all source
distributions. Remember that you only need this mode when you don't want to add
a dependency on versiontools. This will still allow you to use versiontools (in
a limited way) in your setup.py file.

Technically this module defines an improved version of one of
distutils.dist.DistributionMetadata class and monkey-patches distutils to use
it. To retain backward compatibility the new feature is only active when a
special version string is passed to the setup() call.
"""

__version__ = (1, 0, 0, "final", 0)

import distutils.dist
import distutils.errors


class VersiontoolsEnchancedDistributionMetadata(distutils.dist.DistributionMetadata):
    """
    A subclass of distutils.dist.DistributionMetadata that uses versiontools

    Typically you would not instantiate this class directly. It is constructed
    by distutils.dist.Distribution.__init__() method. Since there is no other
    way to do it, this module monkey-patches distutils to override the original
    version of DistributionMetadata
    """

    # Reference to the original class. This is only required because distutils
    # was created before the introduction of new-style classes to python.
    __base = distutils.dist.DistributionMetadata

    def get_version(self): 
        """
        Get distribution version.

        This method is enhanced compared to original distutils implementation.
        If the version string is set to a special value then instead of using
        the actual value the real version is obtained by querying versiontools.

        If versiontools package is not installed then the version is obtained
        from the standard section of the ``PKG-INFO`` file. This file is
        automatically created by any source distribution. This method is less
        useful as it cannot take advantage of version control information that
        is automatically loaded by versiontools. It has the advantage of not
        requiring versiontools installation and that it does not depend on
        ``setup_requires`` feature of ``setuptools``.
        """
        if (self.name is not None and self.version is not None
            and self.version.startswith(":versiontools:")):
            return (self.__get_live_version() or self.__get_frozen_version()
                    or self.__fail_to_get_any_version())
        else:
            return self.__base.get_version(self)

    def __get_live_version(self):
        """
        Get a live version string using versiontools
        """
        try:
            import versiontools
        except ImportError:
            return None
        else:
            return str(versiontools.Version.from_expression(self.name))

    def __get_frozen_version(self):
        """
        Get a fixed version string using an existing PKG-INFO file
        """
        try:
            return self.__base("PKG-INFO").version
        except IOError:
            return None

    def __fail_to_get_any_version(self):
        """
        Raise an informative exception
        """
        raise SystemExit(
"""This package requires versiontools for development or testing.

See http://versiontools.readthedocs.org/ for more information about
what versiontools is and why it is useful.

To install versiontools now please run:
    $ pip install versiontools

Note: versiontools works best when you have additional modules for
integrating with your preferred version control system. Refer to
the documentation for a full list of required modules.""")


# If DistributionMetadata is not a subclass of
# VersiontoolsEnhancedDistributionMetadata then monkey patch it. This should
# prevent a (odd) case of multiple imports of this module.
if not issubclass(
    distutils.dist.DistributionMetadata,
    VersiontoolsEnchancedDistributionMetadata):
    distutils.dist.DistributionMetadata = VersiontoolsEnchancedDistributionMetadata
